"""Configurator classes to implement SMAC2 in Sparkle."""

from __future__ import annotations
from pathlib import Path
import shutil
import math
import random

import pandas as pd

from runrunner import Runner, Run

from sparkle.tools.parameters import PCSConvention
from sparkle.configurator.configurator import Configurator, ConfigurationScenario
from sparkle.solver import Solver
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.instance import InstanceSet, Instance_Set
from sparkle.types import SparkleObjective, resolve_objective


class SMAC2(Configurator):
    """Class for SMAC2 (Java) configurator."""

    configurator_path = Path(__file__).parent.resolve() / "SMAC2"
    configurator_executable = configurator_path / "smac"
    configurator_target = configurator_path / "smac2_target_algorithm.py"

    full_name = "Sequential Model-based Algorithm Configuration"
    version = "2.10.03"

    def __init__(self: SMAC2) -> None:
        """Returns the SMAC2 configurator, Java SMAC V2.10.03."""
        return super().__init__(multi_objective_support=False)

    @property
    def name(self: SMAC2) -> str:
        """Returns the name of the configurator."""
        return SMAC2.__name__

    @staticmethod
    def scenario_class() -> ConfigurationScenario:
        """Returns the SMAC2 scenario class."""
        return SMAC2Scenario

    @staticmethod
    def check_requirements(verbose: bool = False) -> bool:
        """Check that SMAC2 is installed."""
        import warnings

        if no_java := shutil.which("java") is None:
            if verbose:
                warnings.warn(
                    "SMAC2 requires Java 1.8.0_402, but Java is not installed. "
                    "Please ensure Java is installed."
                )
        if no_smac := not SMAC2.configurator_executable.exists():
            if verbose:
                warnings.warn(
                    "SMAC2 executable not found. Please ensure SMAC2 is installed "
                    f"in the expected Path ({SMAC2.configurator_path})."
                )
        return not (no_java or no_smac)

    @staticmethod
    def download_requirements(
        smac2_zip_url: str = "https://github.com/ADA-research/Sparkle/raw/refs/heads/"
        "development/Resources/Configurators/SMAC2-v2.10.03.zip",
    ) -> None:
        """Download SMAC2."""
        if SMAC2.configurator_executable.exists():
            return  # Already installed
        from urllib.request import urlopen
        import zipfile
        import io

        r = urlopen(smac2_zip_url, timeout=60)
        z = zipfile.ZipFile(io.BytesIO(r.read()))
        z.extractall(SMAC2.configurator_path)
        # Ensure execution rights
        SMAC2.configurator_executable.chmod(0o755)

    def configure(
        self: SMAC2,
        scenario: SMAC2Scenario,
        data_target: PerformanceDataFrame,
        validate_after: bool = True,
        sbatch_options: list[str] = [],
        slurm_prepend: str | list[str] | Path = None,
        num_parallel_jobs: int = None,
        base_dir: Path = None,
        run_on: Runner = Runner.SLURM,
    ) -> list[Run]:
        """Start configuration job.

        Args:
            scenario: ConfigurationScenario object
            data_target: PerformanceDataFrame where to store the found configurations
            validate_after: Whether the configurations should be validated on the
                train set afterwards.
            sbatch_options: List of slurm batch options to use.
            slurm_prepend: Slurm script to prepend to the sbatch.
            num_parallel_jobs: The maximum number of jobs to run parallel.
            base_dir: The path where the sbatch scripts will be created for Slurm.
            run_on: On which platform to run the jobs. Default: Slurm.

        Returns:
            A RunRunner Run object.
        """
        scenario.create_scenario()
        configuration_ids = scenario.configuration_ids

        # The maximum seed size for SMAC2 is 999 999 999
        seeds = [random.randint(0, 10**9 - 1) for _ in range(scenario.number_of_runs)]
        output = [
            f"{(scenario.results_directory).absolute()}/"
            f"{scenario.name}_{config_id}_smac.txt"
            for config_id in configuration_ids
        ]
        cmds = [
            f"python3 {Configurator.configurator_cli_path.absolute()} "
            f"{SMAC2.__name__} {output_file} {data_target.csv_filepath} "
            f"{scenario.scenario_file_path} {configuration_id} "
            f"{SMAC2.configurator_executable.absolute()} "
            f"--scenario-file {scenario.scenario_file_path} "
            f"--seed {seed} "
            for output_file, configuration_id, seed in zip(
                output, configuration_ids, seeds
            )
        ]
        if num_parallel_jobs is not None:
            num_parallel_jobs = max(num_parallel_jobs, len(cmds))
        return super().configure(
            configuration_commands=cmds,
            data_target=data_target,
            output=output,
            num_parallel_jobs=num_parallel_jobs,
            scenario=scenario,
            configuration_ids=configuration_ids,
            validate_after=validate_after,
            sbatch_options=sbatch_options,
            slurm_prepend=slurm_prepend,
            base_dir=base_dir,
            run_on=run_on,
        )

    @staticmethod
    def organise_output(
        output_source: Path,
        output_target: Path,
        scenario: SMAC2Scenario,
        configuration_id: str,
    ) -> None | dict:
        """Retrieves configuration from SMAC file and places them in output."""
        call_key = SMAC2.configurator_target.name
        # Last line describing a call is the best found configuration
        for line in reversed(output_source.open("r").readlines()):
            if call_key in line:
                call_str = line.split(call_key, maxsplit=1)[1].strip()
                # The Configuration appears after the first 7 arguments
                configuration = call_str.split(" ", 8)[-1]
                break
        configuration = Solver.config_str_to_dict(configuration)
        configuration["configuration_id"] = configuration_id
        return Configurator.save_configuration(
            scenario, configuration_id, configuration, output_target
        )

    @staticmethod
    def get_smac_run_obj(objective: SparkleObjective) -> str:
        """Return the SMAC run objective based on the Performance Measure.

        Returns:
            A string that represents the run objective set in the settings.
        """
        if objective.time:
            return "RUNTIME"
        return "QUALITY"

    def get_status_from_logs(self: SMAC2, base_dir: Path) -> None:
        """Method to scan the log files of the configurator for warnings."""
        if not base_dir.exists():
            return
        print(
            f"Checking the log files of configurator {type(self).__name__} for "
            "warnings..."
        )
        scenarios = [f for f in base_dir.iterdir() if f.is_dir()]
        for scenario in scenarios:
            log_dir = (
                scenario / "outdir_train_configuration" / (scenario.name + "_scenario")
            )
            # Collect all non empty log files paths
            warn_files = [f for f in log_dir.glob("log-warn*") if f.stat().st_size > 0]
            if len(warn_files) > 0:
                print(
                    f"Scenario {scenario.name} has {len(warn_files)} warning(s), see "
                    "the following log file(s) for more information:"
                )
                for log_file in warn_files:
                    print(f"\t-{log_file}")
            else:
                print(f"Scenario {scenario.name} has no warnings.")


class SMAC2Scenario(ConfigurationScenario):
    """Class to handle SMAC2 configuration scenarios."""

    def __init__(
        self: SMAC2Scenario,
        solver: Solver,
        instance_set: InstanceSet,
        sparkle_objectives: list[SparkleObjective],
        number_of_runs: int,
        parent_directory: Path,
        solver_calls: int = None,
        max_iterations: int = None,
        cpu_time: int = None,
        wallclock_time: int = None,
        solver_cutoff_time: int = None,
        target_cutoff_length: str = None,
        cli_cores: int = None,
        use_cpu_time_in_tunertime: bool = None,
        feature_data: FeatureDataFrame | Path = None,
        timestamp: str = None,
    ) -> None:
        """Initialize scenario paths and names.

        Args:
            solver: Solver that should be configured.
            instance_set: Instances object for the scenario.
            sparkle_objectives: SparkleObjectives used for each run of the configuration.
                Will be simplified to the first objective.
            number_of_runs: The number of configurator runs to perform
                for configuring the solver.
            parent_directory: Directory in which the scenario should be created.
            solver_calls: The number of times the solver is called for each
                configuration run
            max_iterations: The maximum number of iterations allowed for each
                configuration run. [iteration-limit, numIterations, numberOfIterations]
            cpu_time: The time budget allocated for each configuration run. (cpu)
            wallclock_time: The time budget allocated for each configuration run.
                (wallclock)
            solver_cutoff_time: The maximum time allowed for each solver call run during
                configuration.
            target_cutoff_length: A domain specific measure of when the algorithm
                should consider itself done.
            cli_cores: int
                The number of cores to use to execute runs. Defaults in SMAC2 to 1.
            use_cpu_time_in_tunertime: Whether to calculate SMAC2's own used time for
                budget deduction. Defaults in SMAC2 to True.
            feature_data: If features are used, this contains the feature data.
                If it is a FeatureDataFrame, will convert values to SMAC2 format.
                If it is a Path, will pass the path to SMAC2.
                Defaults to None.
            timestamp: An optional timestamp for the directory name.
        """
        super().__init__(
            solver,
            instance_set,
            sparkle_objectives,
            number_of_runs,
            parent_directory,
            timestamp,
        )
        self.solver = solver
        self.instance_set = instance_set

        self.sparkle_objective = sparkle_objectives[0]
        self.solver_calls = solver_calls
        self.cpu_time = cpu_time
        self.wallclock_time = wallclock_time
        self.solver_cutoff_time = solver_cutoff_time
        self.cutoff_length = target_cutoff_length
        self.max_iterations = max_iterations
        self.cli_cores = cli_cores
        self.use_cpu_time_in_tunertime = use_cpu_time_in_tunertime

        self.feature_data = feature_data
        self._feature_file_path = None
        if self.feature_data:
            if isinstance(self.feature_data, FeatureDataFrame):
                # Convert feature data to SMAC2 format
                data_dict = {}
                for instance in self.instance_set.instance_paths:
                    data_dict[str(instance)] = feature_data.get_instance(str(instance))

                self.feature_data = pd.DataFrame.from_dict(
                    data_dict,
                    orient="index",
                    columns=[
                        f"Feature{index + 1}"
                        for index in range(feature_data.num_features)
                    ],
                )

                def map_nan(x: str) -> int:
                    """Map non-numeric values with -512 (Pre-defined by SMAC2)."""
                    if math.isnan(x):
                        return -512.0
                    try:
                        return float(x)
                    except Exception:
                        return -512.0

                self.feature_data = self.feature_data.map(map_nan)
            elif isinstance(self.feature_data, Path):  # Read from Path
                self._feature_file_path = feature_data
                self.feature_data = pd.read_csv(self.feature_file_path, index_col=0)
            else:
                print(
                    f"WARNING: Feature data is of type {type(feature_data)}. "
                    "Expected FeatureDataFrame or Path."
                )

    @property
    def instance_file_path(self: SMAC2Scenario) -> Path:
        """Return the path of the instance file."""
        if self.directory:
            return self.directory / f"{self.instance_set.name}.txt"
        return None

    @property
    def outdir_train(self: SMAC2Scenario) -> Path:
        """Return the path of the train out directory."""
        # SMAC2 Specific directory
        if self.directory:
            return self.directory / "outdir_train_configuration"
        return None

    @property
    def feature_file_path(self: SMAC2Scenario) -> Path:
        """Return the path of the feature file."""
        if self._feature_file_path:
            return self._feature_file_path
        elif self.directory:
            return self.directory / f"{self.instance_set.name}_features.csv"
        else:
            return None

    @property
    def configurator(self: SMAC2Scenario) -> SMAC2:
        """Return the type of configurator the scenario belongs to."""
        return SMAC2

    def create_scenario(self: SMAC2Scenario) -> None:
        """Create scenario with solver and instances in the parent directory.

        This prepares all the necessary subdirectories related to configuration.

        Args:
            parent_directory: Directory in which the scenario should be created.
        """
        super().create_scenario()
        self.outdir_train.mkdir()
        self._prepare_instances()

        if self.feature_data is not None:
            self._create_feature_file()

        self.create_scenario_file()

    def create_scenario_file(
        self: SMAC2Scenario,
        configurator_target: Path = SMAC2.configurator_target,
        pcs_port: PCSConvention = PCSConvention.SMAC,
    ) -> Path:
        """Create a file with the configuration scenario.

        Writes supplementary information to the target algorithm (algo =) as:
        algo = {configurator_target} {solver_directory} {sparkle_objective}
        """
        with self.scenario_file_path.open("w") as file:
            file.write(
                f"algo = {configurator_target.absolute()} "
                f"{self.solver.directory} {self.tmp} {self.sparkle_objective} \n"
                f"deterministic = {1 if self.solver.deterministic else 0}\n"
                f"run_obj = {self._get_performance_measure()}\n"
                f"cutoffTime = {self.solver_cutoff_time}\n"
                f"paramfile = {self.solver.get_pcs_file(pcs_port)}\n"
                f"outdir = {self.outdir_train}\n"
                f"instance_file = {self.instance_file_path}\n"
                f"test_instance_file = {self.instance_file_path}\n"
            )
            if self.cutoff_length is not None:
                file.write(f"cutoff_length = {self.cutoff_length}\n")
            if self.max_iterations is not None:
                file.write(f"iteration-limit = {self.max_iterations}\n")
            if self.wallclock_time is not None:
                file.write(f"wallclock-limit = {self.wallclock_time}\n")
            if self.cpu_time is not None:
                file.write(f"cputime-limit = {self.cpu_time}\n")
            if self.solver_calls is not None:
                file.write(f"runcount-limit = {self.solver_calls}\n")
            if self.cli_cores is not None:
                file.write(f"cli-cores = {self.cli_cores}")
            if self.feature_data is not None:
                file.write(f"feature_file = {self.feature_file_path}\n")
            if self.use_cpu_time_in_tunertime is not None:
                file.write(
                    f"use-cpu-time-in-tunertime = {self.use_cpu_time_in_tunertime}\n"
                )
            # We don't let SMAC do the validation
            file.write("validation = false" + "\n")
        return self.scenario_file_path

    def _prepare_instances(self: SMAC2Scenario) -> None:
        """Create instance list file without instance specifics."""
        self.instance_file_path.parent.mkdir(exist_ok=True, parents=True)
        with self.instance_file_path.open("w+") as file:
            for instance_path in self.instance_set._instance_paths:
                file.write(f"{instance_path}\n")

    def _create_feature_file(self: SMAC2Scenario) -> None:
        """Create CSV file from feature data."""
        self.feature_data.to_csv(self.feature_file_path, index_label="INSTANCE_NAME")

    def _get_performance_measure(self: SMAC2Scenario) -> str:
        """Retrieve the performance measure of the SparkleObjective.

        Returns:
            Performance measure of the sparkle objective
        """
        if self.sparkle_objective.time:
            return "RUNTIME"
        return "QUALITY"

    def serialise(self: SMAC2Scenario) -> dict:
        """Transform ConfigurationScenario to dictionary format."""
        return {
            "number_of_runs": self.number_of_runs,
            "solver_calls": self.solver_calls,
            "cpu_time": self.cpu_time,
            "wallclock_time": self.wallclock_time,
            "solver_cutoff_time": self.solver_cutoff_time,
            "cutoff_length": self.cutoff_length,
            "max_iterations": self.max_iterations,
            "sparkle_objective": self.sparkle_objective.name,
            "feature_data": str(self.feature_file_path),
            "use_cpu_time_in_tunertime": self.use_cpu_time_in_tunertime,
        }

    @staticmethod
    def from_file(scenario_file: Path) -> SMAC2Scenario:
        """Reads scenario file and initalises SMAC2Scenario."""
        config = {
            keyvalue[0]: keyvalue[1]
            for keyvalue in (
                line.strip().split(" = ", maxsplit=1)
                for line in scenario_file.open().readlines()
                if line.strip() != ""
            )
        }

        # Collect relevant settings
        cpu_time = int(config["cpu_time"]) if "cpu_time" in config else None
        wallclock_limit = (
            int(config["wallclock-limit"]) if "wallclock-limit" in config else None
        )
        solver_calls = (
            int(config["runcount-limit"]) if "runcount-limit" in config else None
        )
        max_iterations = (
            int(config["iteration-limit"]) if "iteration-limit" in config else None
        )
        use_cpu_time_in_tunertime = (
            config["use-cputime-in-tunertime"]
            if "use-cputime-in-tunertime" in config
            else None
        )
        cli_cores = config["cli-cores"] if "cli-cores" in config else None

        _, solver_path, _, objective_str = config["algo"].split(" ")
        objective = resolve_objective(objective_str)
        solver = Solver(Path(solver_path.strip()))
        # Extract the instance set from the instance file
        instance_file_path = Path(config["instance_file"])
        instance_set_path = Path(instance_file_path.open().readline().strip()).parent
        instance_set = Instance_Set(Path(instance_set_path))
        results_folder = scenario_file.parent / "results"
        state_run_dirs = [p for p in results_folder.iterdir() if p.is_file()]
        number_of_runs = len(state_run_dirs)
        feature_data_path = None
        if "feature_file" in config:
            feature_data_path = Path(config["feature_file"])
        # Get the timestamp from the scenario dir name
        timestamp = scenario_file.parent.name.split("_")[-1]
        return SMAC2Scenario(
            solver,
            instance_set,
            [objective],
            number_of_runs,
            instance_file_path.parent.parent,
            solver_calls,
            max_iterations,
            cpu_time,
            wallclock_limit,
            int(config["cutoffTime"]),
            config["cutoff_length"],
            cli_cores,
            use_cpu_time_in_tunertime,
            feature_data_path,
            timestamp,
        )



#!/usr/bin/env python3
"""Sparkle command to add a feature extractor to the Sparkle platform."""

import os
import stat
import sys
import shutil
import argparse
from pathlib import Path

from sparkle.platform import file_help as sfh
from sparkle.CLI.help import global_variables as gv
from sparkle.structures import FeatureDataFrame
from sparkle.CLI.help import logging as sl
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as ac
from sparkle.selector import Extractor


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    # Define command line arguments
    parser = argparse.ArgumentParser(
        description="Add a feature extractor to the platform."
    )
    parser.add_argument(
        *ac.ExtractorPathArgument.names, **ac.ExtractorPathArgument.kwargs
    )
    parser.add_argument(
        *ac.NicknameFeatureExtractorArgument.names,
        **ac.NicknameFeatureExtractorArgument.kwargs,
    )
    parser.add_argument(*ac.NoCopyArgument.names, **ac.NoCopyArgument.kwargs)
    return parser


def main(argv: list[str]) -> None:
    """Main function of the add feature extractor command."""
    # Log command call
    sl.log_command(sys.argv, gv.settings().random_state)
    check_for_initialise()

    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)

    extractor_source = Path(args.extractor_path)
    if not extractor_source.exists():
        print(f'Feature extractor path "{extractor_source}" does not exist!')
        sys.exit(-1)

    nickname_str = args.nickname

    # Start add feature extractor
    extractor_target_path = gv.settings().DEFAULT_extractor_dir / extractor_source.name

    if extractor_target_path.exists():
        print(
            f"Feature extractor {extractor_source.name} already exists! "
            "Can not add feature extractor."
        )
        sys.exit(-1)

    if args.no_copy:
        print(
            f"Creating symbolic link from {extractor_source} "
            f"to {extractor_target_path}..."
        )
        extractor_target_path.symlink_to(extractor_source.absolute())
    else:
        print(f"Copying feature extractor {extractor_source.name} ...")
        extractor_target_path.mkdir()
        shutil.copytree(extractor_source, extractor_target_path, dirs_exist_ok=True)

    # Check execution permissions for wrapper
    extractor_wrapper = extractor_target_path / Extractor.wrapper
    if not extractor_wrapper.is_file() or not os.access(extractor_wrapper, os.X_OK):
        print(
            f"The file {extractor_wrapper} does not exist or is \
              not executable."
        )
        sys.exit(-1)

    extractor = Extractor(extractor_target_path)

    # Add RunSolver executable to the solver
    runsolver_path = gv.settings().DEFAULT_runsolver_exec
    if runsolver_path.name in [file.name for file in extractor_target_path.iterdir()]:
        print(
            "Warning! RunSolver executable detected in Extractor "
            f"{extractor.name}. This will be replaced with "
            f"Sparkle's version of RunSolver. ({runsolver_path})"
        )

    if runsolver_path.exists():
        runsolver_target = extractor.directory / runsolver_path.name
        shutil.copyfile(runsolver_path, runsolver_target)
        runsolver_target.chmod(runsolver_target.stat().st_mode | stat.S_IEXEC)
    else:
        print("Warning! RunSolver does not exists. Falling back to PyRunSolver.")

    # Get the extractor features groups and names from the wrapper
    feature_dataframe = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)
    feature_dataframe.add_extractor(extractor.name, extractor.features)
    feature_dataframe.save_csv()

    print(f"Adding feature extractor {extractor_target_path.name} done!")

    if nickname_str is not None:
        sfh.add_remove_platform_item(
            extractor_target_path,
            gv.extractor_nickname_list_path,
            gv.file_storage_data_mapping[gv.extractor_nickname_list_path],
            key=nickname_str,
        )

    # Write used settings to file
    gv.settings().write_used_settings()
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Execute Feature Extractor for an instance, write features to FeatureDataFrame."""

import argparse
from pathlib import Path
from filelock import FileLock

from sparkle.structures import FeatureDataFrame
from sparkle.selector import Extractor


if __name__ == "__main__":
    # Define command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--extractor", required=True, type=Path, help="path to feature extractor"
    )
    parser.add_argument(
        "--instance",
        required=True,
        type=Path,
        nargs="+",
        help="path to instance file(s) to run on",
    )
    parser.add_argument(
        "--feature-csv", required=True, type=Path, help="path to feature data CSV file"
    )
    parser.add_argument(
        "--cutoff",
        required=True,
        type=str,
        help="the maximum CPU time for the extractor.",
    )
    parser.add_argument(
        "--feature-group",
        required=False,
        type=str,
        help="the group of features to compute, if available for the "
        "extractor. If not available or provided, all groups will"
        " be computed.",
    )
    parser.add_argument(
        "--log-dir", type=Path, required=True, help="path to the log directory"
    )
    args = parser.parse_args()

    # Process command line arguments
    log_dir = args.log_dir

    # Instance agument is a list to allow for multifile instances
    instance_path: list[Path] = args.instance
    instance_name = instance_path[0].stem
    extractor_path = args.extractor
    feature_data_csv_path = args.feature_csv
    cutoff_extractor = args.cutoff

    # Ensure stringifcation of path objects
    if isinstance(instance_path, list):
        instance_list = [str(filepath) for filepath in instance_path]
    else:
        instance_list = [str(instance_path)]

    extractor = Extractor(extractor_path)
    features = extractor.run(
        instance_list,
        feature_group=args.feature_group,
        cutoff_time=cutoff_extractor,
        log_dir=log_dir,
    )

    # Now that we have our result, we write it to the FeatureDataCSV with a FileLock
    lock = FileLock(f"{feature_data_csv_path}.lock")
    if features is not None:
        print(f"Writing features to CSV: {instance_name}, {extractor_path.name}")
        with lock.acquire(timeout=60):
            feature_data = FeatureDataFrame(feature_data_csv_path)
            instance_key = (
                instance_name
                if instance_name in feature_data.instances
                else str(instance_path[0].with_suffix(""))
            )
            for feature_group, feature_name, value in features:
                feature_data.set_value(
                    instance_key,
                    extractor_path.name,
                    feature_group,
                    feature_name,
                    float(value),
                )
            feature_data.save_csv()
        lock.release()
    else:
        print(
            "EXCEPTION during retrieving extractor results.\n"
            f"****** WARNING: Feature vector computation on instance {instance_path}"
            " failed! ******"
        )



"""Command to help users check if their input solvers, datasets etc. are correct."""

import sys
import os
import argparse

from runrunner import Runner

from sparkle.solver import Solver
from sparkle.instance import Instance_Set, InstanceSet
from sparkle.selector.extractor import Extractor
from sparkle.platform import Settings
from sparkle.types import SolverStatus

from sparkle.CLI.help import logging as sl
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help import global_variables as gv


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Command to help users check if their input solvers, instance sets "
        "or feature extractors are readable by Sparkle. Specify a path to "
        "an instance to test calling the wrapper."
    )
    parser.add_argument(*ac.CheckTypeArgument.names, **ac.CheckTypeArgument.kwargs)
    parser.add_argument(*ac.CheckPathArgument.names, **ac.CheckPathArgument.kwargs)
    parser.add_argument(*ac.InstancePathOptional.names, **ac.InstancePathOptional.kwargs)
    parser.add_argument(*ac.SeedArgument.names, **ac.SeedArgument.kwargs)
    # Settings arguments
    parser.add_argument(
        *Settings.OPTION_solver_cutoff_time.args,
        **Settings.OPTION_solver_cutoff_time.kwargs,
    )
    return parser


def main(argv: list[str]) -> None:
    """Main entry point of the Check command."""
    parser = parser_function()
    args = parser.parse_args(argv)
    settings = gv.settings(args)

    # Log command call
    sl.log_command(sys.argv, settings.random_state)

    type_map = {
        "extractor": Extractor,
        "feature-extractor": Extractor,
        "solver": Solver,
        "instance-set": Instance_Set,
        "Extractor": Extractor,
        "Feature-Extractor": Extractor,
        "Instance-Set": Instance_Set,
        "Solver": Solver,
        "FeatureExtractor": Extractor,
        "InstanceSet": Instance_Set,
    }
    type = type_map[args.type]
    print(f"Checking {type.__name__} in directory {args.path} ...")
    object = type(args.path)
    print("Resolved to:")
    print(object.__repr__())

    # Conduct object specific tests
    if isinstance(object, Solver):
        if object.pcs_file:
            print()
            print(object.get_configuration_space())
        if not os.access(object.wrapper_file, os.X_OK):
            print(
                f"Wrapper {object.wrapper_file} is not executable!"
                f"Check that wrapper execution rights are set for all."
            )
            sys.exit(-1)
        if args.instance_path:  # Instance to test with
            object._runsolver_exec = settings.DEFAULT_runsolver_exec
            if not object.runsolver_exec.exists():
                print(
                    f"Runsolver {object.runsolver_exec} not found. "
                    "Checking without Runsolver currently not supported."
                )
            else:
                # Test the wrapper with a dummy call
                print(f"\nTesting Wrapper {object.wrapper} ...")
                # Patchfix runsolver
                objectives = settings.objectives if settings.objectives else []
                cutoff = (
                    settings.solver_cutoff_time if settings.solver_cutoff_time else 5
                )
                configuration = {}
                if object.pcs_file:
                    print("\nSample Configuration:")
                    sample_conf = object.get_configuration_space().sample_configuration()
                    print(sample_conf)
                    configuration = dict(sample_conf)
                result = object.run(
                    instances=args.instance_path,
                    objectives=objectives,
                    seed=42,  # Dummy seed
                    cutoff_time=cutoff,
                    configuration=configuration,
                    log_dir=sl.caller_log_dir,
                    run_on=Runner.LOCAL,
                )
                print("Result:")
                for obj in objectives:  # Check objectives
                    if obj.name not in result:
                        print(f"\tSolver output is missing objective {obj.name}")
                    else:
                        print(f"\t{obj.name}: {result[obj.name]}")
                status_key = [key for key in result.keys() if key.startswith("status")][
                    0
                ]
                if result[status_key] == SolverStatus.UNKNOWN:
                    print(
                        f"Sparkle was unable to process {obj.name} output. "
                        "Check that your wrapper is able to handle KILL SIGNALS. "
                        "It is important to always communicate back to Sparkle "
                        "on regular exit and termination signals for stability."
                    )
                    sys.exit(-1)
    elif isinstance(object, Extractor):
        if args.instance_path:  # Test on an instance
            # Patchfix runsolver
            object.runsolver_exec = settings.DEFAULT_runsolver_exec
            if not object.runsolver_exec.exists():
                print(
                    f"Runsolver {object.runsolver_exec} not found. "
                    "Checking without Runsolver currently not supported."
                )
            else:
                print(f"\nTesting Wrapper {object.wrapper} ...")
                # cutoff = args.cutoff_time if args.cutoff_time else 20  # Short call
                result = object.run(
                    instance=args.instance_path, log_dir=sl.caller_log_dir
                )
                # Print feature results
                print("Feature values:")
                for f_group, f_name, f_value in result:
                    print(f"\t[{f_group}] {f_name}: {f_value}")
        else:
            print("Feature names:")
            for f_group, fname in object.features:
                print(f"\t{f_group}: {fname}")
    elif isinstance(object, InstanceSet):
        print("\nList of instances:")
        for i_name, i_path in zip(object.instance_names, object.instance_paths):
            print(f"\t{i_name}: {i_path}")

    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



"""Classes for verifying solutions."""

from __future__ import annotations
from pathlib import Path
import subprocess

from sparkle.types import SolverStatus


class SolutionVerifier:
    """Solution verifier base class."""

    def verify(
        self: SolutionVerifier, instance: Path, output: dict, solver_call: list[str]
    ) -> SolverStatus:
        """Verify the solution."""
        raise NotImplementedError


class SATVerifier(SolutionVerifier):
    """Class to handle the SAT verifier."""

    sat_verifier_path = (
        Path(__file__).parent.parent / "Components/Sparkle-SAT-verifier/SAT"
    )

    def __str__(self: SATVerifier) -> str:
        """Return the name of the SAT verifier."""
        return SATVerifier.__name__

    @staticmethod
    def verify(instance: Path, output: dict, solver_call: list[str]) -> SolverStatus:
        """Run a SAT verifier and return its status."""
        if SolverStatus(output["status"]) == SolverStatus.TIMEOUT:
            return SolverStatus.TIMEOUT
        raw_result = Path([s for s in solver_call if s.endswith(".rawres")][0])
        return SATVerifier.call_sat_raw_result(instance, raw_result)

    @staticmethod
    def sat_verify_output(sat_output: str) -> SolverStatus:
        """Return the status of the SAT verifier.

        Four statuses are possible: "SAT", "UNSAT", "WRONG", "UNKNOWN"
        """
        return_code = int(sat_output.splitlines()[-1])
        if return_code == 11:  # SAT code
            return SolverStatus.SAT
        if return_code == 10:  # UNSAT code
            return SolverStatus.UNSAT
        # Wrong code OR Unknown code
        if return_code == 0 and "Wrong solution." in sat_output:
            return SolverStatus.WRONG
        # Should not occur
        return SolverStatus.UNKNOWN

    @staticmethod
    def call_sat_raw_result(instance: Path, raw_result: Path) -> SolverStatus:
        """Run a SAT verifier to determine correctness of a result.

        Args:
            instance: path to the instance
            raw_result: path to the result to verify

        Returns:
            The status of the solver on the instance
        """
        sat_verify = subprocess.run(
            [SATVerifier.sat_verifier_path, instance, raw_result], capture_output=True
        )
        return SATVerifier.sat_verify_output(sat_verify.stdout.decode())


class SolutionFileVerifier(SolutionVerifier):
    """Class to handle the file verifier."""

    def __init__(self: SolutionFileVerifier, csv_file: Path) -> None:
        """Initialize the verifier by building dictionary from csv.

        Args:
            csv_file: path to the csv file. Requires lines to be of the form:
                instance,objective,solution
        """
        self.csv_file = csv_file
        lines = [
            line.split(",", maxsplit=2)
            for line in self.csv_file.read_text().splitlines()
        ]
        self.solutions = {
            instance: (objective, solution) for instance, objective, solution in lines
        }

    def __str__(self: SATVerifier) -> str:
        """Return the name of the SAT verifier."""
        return SolutionFileVerifier.__name__

    def verify(
        self: SolutionFileVerifier,
        instance: Path,
        solver_output: dict[str, object],
        solver_call: list[str],
    ) -> SolverStatus:
        """Verify the solution.

        Args:
            instance: instance to verify, solution found by instance name as key
            solver_output: outcome of the solver to verify
            solver_call: list of solver calls

        Returns:
            The status of the solver on the instance
        """
        # If the solver ran out of time, we cannot verify a solution
        if SolverStatus(solver_output["status"]) == SolverStatus.TIMEOUT:
            return SolverStatus.TIMEOUT
        instance = instance.name
        if instance not in self.solutions:
            return SolverStatus.UNKNOWN

        objective, solution = self.solutions[instance]
        if objective not in solver_output:
            return SolverStatus.UNKNOWN

        outcome = solver_output[objective]
        if solution in SolverStatus._value2member_map_:  # SAT type status
            solution, outcome = SolverStatus(solution), SolverStatus(outcome)
            if solution == SolverStatus.UNKNOWN or solution == outcome:
                # Verify that the presented solution is correct
                return SATVerifier.verify(instance, solver_output, solver_call)
            else:
                return SolverStatus.WRONG

        if solution != outcome:
            return SolverStatus.WRONG
        return SolverStatus.SUCCESS


# Define a mapping so we can translate between names and classes
mapping = {
    SATVerifier.__name__: SATVerifier,
    SolutionFileVerifier.__name__: SolutionFileVerifier,
}



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Handles SMAC2 calls passing to sparkle solver wrappers."""

import sys
from pathlib import Path

from runrunner import Runner

from sparkle.solver import Solver
from sparkle.types import resolve_objective


if __name__ == "__main__":
    # Incoming call from SMAC:
    # Translate input to Solver object input
    argsiter = iter(sys.argv[9:])
    args = zip(argsiter, argsiter)
    configuration = {arg.strip("-"): val for arg, val in args}
    # Args 1-8 conditions of the run, the rest are configurations for the solver
    # [Solver_dir, Solver_Log_dir, SparkleObjective, instance, specifics,
    #  cutoff_time, runlength, seed]
    solver_dir = Path(sys.argv[1])
    solver_log_dir = Path(sys.argv[2])
    objective = resolve_objective(sys.argv[3])
    instance = sys.argv[4]
    cutoff_time = float(sys.argv[6])
    seed = int(sys.argv[8])

    runsolver_binary = solver_dir / "runsolver"
    solver = Solver(solver_dir, runsolver_exec=runsolver_binary)
    # Call Runsolver with the solver configurator wrapper and its arguments
    output = solver.run(
        instances=instance,
        objectives=[objective],
        seed=seed,
        cutoff_time=cutoff_time,
        configuration=configuration,
        log_dir=solver_log_dir,
        run_on=Runner.LOCAL,
    )

    # Return values to SMAC
    # SMAC2 does not accept nan values etc for quality
    quality = "0"
    if not objective.time and objective.name in output.keys():
        quality = float(output[objective.name])
        if not objective.minimise:
            quality = -1 * quality
    print(
        f"Result for SMAC: {output['status']}, {output['cpu_time']}, 0, {quality}, {seed}"
    )



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Run a solver, read/write to performance dataframe."""

import sys
from filelock import FileLock
import argparse
from pathlib import Path
import random
import time

from runrunner import Runner

from sparkle.solver import Solver
from sparkle.types import resolve_objective
from sparkle.structures import PerformanceDataFrame
from sparkle.tools.solver_wrapper_parsing import parse_commandline_dict


def main(argv: list[str]) -> None:
    """Main function of the command."""
    # Define command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--performance-dataframe",
        required=True,
        type=Path,
        help="path to the performance dataframe",
    )
    parser.add_argument("--solver", required=True, type=Path, help="path to solver")
    parser.add_argument(
        "--instance",
        required=True,
        type=Path,
        nargs="+",
        help="path to instance to run on",
    )
    parser.add_argument(
        "--run-index",
        required=True,
        type=int,
        help="run index in the dataframe to set.",
    )
    parser.add_argument(
        "--log-dir", type=Path, required=True, help="path to the log directory"
    )

    # These two arguments should be mutually exclusive
    parser.add_argument(
        "--configuration-id",
        type=str,
        required=False,
        help="configuration id to read from the PerformanceDataFrame.",
    )
    parser.add_argument(
        "--configuration",
        type=str,
        nargs="+",
        required=False,
        help="configuration for the solver",
    )

    parser.add_argument(
        "--seed",
        type=int,
        required=False,
        help="seed to use for the solver. If not provided, generates one.",
    )
    parser.add_argument(
        "--cutoff-time",
        type=int,
        required=False,
        help="the cutoff time for the solver.",
    )
    parser.add_argument(
        "--objectives",
        type=str,
        required=False,
        nargs="+",
        help="The objectives to evaluate to Solver on. If not provided, read from the PerformanceDataFrame.",
    )
    parser.add_argument(
        "--target-objective",
        required=False,
        type=str,
        help="The objective to use to determine the best configuration.",
    )
    parser.add_argument(
        "--best-configuration-instances",
        required=False,
        type=str,
        nargs="+",
        help="If given, will ignore any given configurations, and try to"
        " determine the best found configurations over the given "
        "instances. Uses the 'target-objective' given in the arguments"
        " or the first one given by the dataframe to determine the best"
        "configuration.",
    )
    args = parser.parse_args(argv)
    # Process command line arguments
    log_dir = args.log_dir
    print(f"Running Solver and read/writing results with {args.performance_dataframe}")
    # Resolve possible multi-file instance
    instance_path: list[Path] = args.instance
    # If instance is only one file then we don't need a list
    instance_path = instance_path[0] if len(instance_path) == 1 else instance_path
    instance_name = (
        instance_path.stem if isinstance(instance_path, Path) else instance_path[0].stem
    )
    run_index = args.run_index
    # Ensure stringifcation of path objects
    if isinstance(instance_path, list):
        # Double list because of solver.run
        run_instances = [[str(filepath) for filepath in instance_path]]
    else:
        run_instances = str(instance_path)

    solver = Solver(args.solver)
    # By default, run the default configuration
    config_id = PerformanceDataFrame.default_configuration
    configuration = None
    # If no seed is provided by CLI, generate one
    seed = args.seed if args.seed else random.randint(0, 2**32 - 1)
    # Parse the provided objectives if present
    objectives = (
        [resolve_objective(o) for o in args.objectives] if args.objectives else None
    )

    if args.configuration:  # Configuration provided, override
        configuration = parse_commandline_dict(args.configuration)
        config_id = configuration["configuration_id"]
    elif (
        (
            args.configuration_id
            and args.configuration_id != PerformanceDataFrame.default_configuration
        )
        or args.best_configuration_instances
        or not objectives
    ):  # Read from PerformanceDataFrame, can be slow
        # Desyncronize from other possible jobs writing to the same file
        print(
            "Reading from Performance DataFrame.. "
            f"[{'configuration' if (args.configuration_id or args.best_configuration_instances) else ''} "
            f"{'objectives' if not objectives else ''}]"
        )
        time.sleep(random.random() * 10)
        lock = FileLock(f"{args.performance_dataframe}.lock")  # Lock the file
        with lock.acquire(timeout=600):
            performance_dataframe = PerformanceDataFrame(args.performance_dataframe)

        if not objectives:
            objectives = performance_dataframe.objectives

        if args.best_configuration_instances:  # Determine best configuration
            best_configuration_instances: list[str] = args.best_configuration_instances
            # Get the unique instance names
            best_configuration_instances = list(
                set([Path(instance).stem for instance in best_configuration_instances])
            )
            target_objective = (
                resolve_objective(args.target_objective)
                if args.target_objective
                else objectives[0]
            )
            config_id, _ = performance_dataframe.best_configuration(
                solver=str(args.solver),
                objective=target_objective,
                instances=best_configuration_instances,
            )
            configuration = performance_dataframe.get_full_configuration(
                str(args.solver), config_id
            )

        elif (
            args.configuration_id
        ):  # Read from PerformanceDataFrame the configuration using the ID
            config_id = args.configuration_id
            configuration = performance_dataframe.get_full_configuration(
                str(args.solver), config_id
            )

    print(f"Running Solver {solver} on instance {instance_name} with seed {seed}..")
    solver_output = solver.run(
        run_instances,
        objectives=objectives,
        seed=seed,
        configuration=configuration.copy() if configuration else None,
        cutoff_time=args.cutoff_time,
        log_dir=log_dir,
        run_on=Runner.LOCAL,
    )

    # Prepare the results for the DataFrame for each objective
    result = [
        [solver_output[objective.name] for objective in objectives],
        [seed] * len(objectives),
    ]
    solver_fields = [
        PerformanceDataFrame.column_value,
        PerformanceDataFrame.column_seed,
    ]

    print(f"For Solver/config: {solver}/{config_id}")
    print(f"For index: Instance {instance_name}, Run {args.run_index}, Seed {seed}")
    print("Appending the following objective values:")  # {', '.join(objective_values)}")
    for objective in objectives:
        print(
            f"{objective.name}, {instance_name}, {args.run_index} | {args.solver}, {config_id}: {solver_output[objective.name]}"
        )

    # Desyncronize from other possible jobs writing to the same file
    time.sleep(random.random() * 100)
    # TESTLOG
    # lock = FileLock("test.log.lock")
    # with lock.acquire(timeout=600):
    #     with Path("test.log").open("a") as f:
    #         for objective in objectives:
    #             f.write(f"{objective.name}, {instance_name}, {args.run_index} | {solver} {config_id}: {solver_output[objective.name]}, {seed}\n")

    # Now that we have all the results, we can add them to the performance dataframe
    lock = FileLock(f"{args.performance_dataframe}.lock")  # Lock the file
    with lock.acquire(timeout=600):
        performance_dataframe = PerformanceDataFrame(args.performance_dataframe)
        performance_dataframe.set_value(
            result,
            solver=str(args.solver),
            instance=instance_name,
            configuration=config_id,
            objective=[o.name for o in objectives],
            run=run_index,
            solver_fields=solver_fields,
            append_write_csv=True,  # We do not have to save the PDF here, thanks to this argument
        )


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
"""Command to display the status of the platform."""

import sys
import argparse
from pathlib import Path

from sparkle.structures import FeatureDataFrame, PerformanceDataFrame

from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.platform import Settings


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(description="Display the status of the platform.")
    parser.add_argument(
        *Settings.OPTION_verbosity.args, **Settings.OPTION_verbosity.kwargs
    )
    return parser


def print_objects_list(objects: list[any], type: str, details: bool = False) -> None:
    """Print a list of sparkle objects.

    Args:
        objects: The objects to print
        type: The name of the object type
        details: Indicating if output should be detailed
    """
    print(
        f"Currently Sparkle has {len(objects)} {type}(s)"
        + (":" if details and objects else "")
    )

    if details:
        for i, object in enumerate(objects):
            print(f"\t[{i + 1}] {Path(object).name}")


def print_feature_computation_jobs(
    feature_data_csv: Path, verbose: bool = False
) -> None:
    """Print a list of remaining feature computation jobs.

    Args:
        feature_data_csv: Path to the feature data csv
        verbose: Indicating, if output should be verbose
    """
    if not feature_data_csv.exists():
        print("\nNo feature data found, cannot determine remaining jobs.")

    feature_data = FeatureDataFrame(feature_data_csv)
    jobs = feature_data.remaining_jobs()

    print(
        f"Currently Sparkle has {len(jobs)} remaining feature computation "
        "jobs that need to be performed before creating an algorithm selector"
        + (":" if verbose and jobs else "")
    )

    if verbose:
        for i, job in enumerate(jobs):
            print(
                f"[{i + 1}]: Extractor: {Path(job[1]).name}, Group: {job[2]}, "
                f"Instance: {Path(job[0]).name}"
            )


def print_performance_computation_jobs(
    performance_data: PerformanceDataFrame, verbose: bool = False
) -> None:
    """Print a list of remaining performance computation jobs.

    Args:
        performance_data: The Performance data
        verbose: Indicating, if output should be verbose
    """
    jobs = performance_data.get_job_list()

    print(
        f"Currently Sparkle has {len(jobs)} remaining performance computation"
        " jobs that need to be performed before creating an algorithm selector"
        + (":" if verbose else "")
    )

    if verbose:
        i = 0
        for solver, config, instance, run in jobs:
            print(
                f"[{i + 1}]: Solver: "
                f"{Path(solver).name} ({config}), Instance: "
                f"{Path(instance).name}"
            )
            i += 1


def main(argv: list[str]) -> None:
    """Main function of the status command."""
    # Log command call
    sl.log_command(sys.argv, gv.settings().random_state)
    check_for_initialise()

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)

    performance_data = PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path)

    print("========Sparkle System Status========\n")
    print_objects_list(
        [s for s in gv.settings().DEFAULT_solver_dir.iterdir()],
        "Solver",
        args.verbosity,
    )
    if args.verbosity:
        print()
    print_objects_list(
        [e for e in gv.settings().DEFAULT_extractor_dir.iterdir()],
        "Extractor",
        args.verbosity,
    )
    if args.verbosity:
        print()
    print_objects_list(
        [i for i in gv.settings().DEFAULT_instance_dir.iterdir()],
        "Instance Set",
        args.verbosity,
    )

    print()
    print_feature_computation_jobs(
        gv.settings().DEFAULT_feature_data_path, args.verbosity
    )
    print_performance_computation_jobs(performance_data, args.verbosity)

    if args.verbosity:
        print("\nThe Performance Data overview:")
        print(performance_data)

    # scan configurator log files for warnings
    configurator = gv.settings().configurator
    configurator.get_status_from_logs(
        gv.settings().get_configurator_output_path(configurator)
    )

    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



"""Configurator class to use different algorithm configurators."""

from __future__ import annotations
import re
import shutil
import decimal
from pathlib import Path
from datetime import datetime
from typing import Optional
import random

import runrunner as rrr
from runrunner import Runner, Run

from sparkle.solver import Solver
from sparkle.instance import InstanceSet, Instance_Set
from sparkle.structures import PerformanceDataFrame
from sparkle.types import SparkleObjective


class Configurator:
    """Abstact class to use different configurators like SMAC."""

    configurator_cli_path = Path(__file__).parent.resolve() / "configurator_cli.py"

    full_name = "Configurator Abstract Class"
    version = "NaN"

    def __init__(self: Configurator, multi_objective_support: bool = False) -> None:
        """Initialize Configurator.

        Args:
            multi_objective_support: Whether the configurator supports
                multi objective optimization for solvers.
        """
        self.multiobjective = multi_objective_support

    @property
    def name(self: Configurator) -> str:
        """Return the name of the configurator."""
        return self.__class__.__name__

    @staticmethod
    def scenario_class() -> ConfigurationScenario:
        """Return the scenario class of the configurator."""
        return ConfigurationScenario

    @staticmethod
    def check_requirements(verbose: bool = False) -> bool:
        """Check if the configurator is installed."""
        raise NotImplementedError

    @staticmethod
    def download_requirements() -> None:
        """Download the configurator."""
        raise NotImplementedError

    def configure(
        self: Configurator,
        configuration_commands: list[str],
        data_target: PerformanceDataFrame,
        output: Path,
        scenario: ConfigurationScenario,
        configuration_ids: list[str] = None,
        validate_after: bool = True,
        sbatch_options: list[str] = None,
        slurm_prepend: str | list[str] | Path = None,
        num_parallel_jobs: int = None,
        base_dir: Path = None,
        run_on: Runner = Runner.SLURM,
    ) -> Run:
        """Start configuration job.

        This method is shared by the configurators and should be called by the
        implementation/subclass of the configurator.

        Args:
            configuration_commands: List of configurator commands to execute
            data_target: Performance data to store the results.
            output: Output directory.
            scenario: ConfigurationScenario to execute.
            configuration_ids: List of configuration ids that are to be created
            validate_after: Whether the configurations should be validated
            sbatch_options: List of slurm batch options to use
            slurm_prepend: Slurm script to prepend to the sbatch
            num_parallel_jobs: The maximum number of jobs to run in parallel
            base_dir: The base_dir of RunRunner where the sbatch scripts will be placed
            run_on: On which platform to run the jobs. Default: Slurm.

        Returns:
            A RunRunner Run object.
        """
        if not self.check_requirements(verbose=True):
            raise RuntimeError(
                f"{self.name} is not installed. Please install {self.name} and try again."
            )
        # Add the configuration IDs to the dataframe with empty configurations
        data_target.add_configuration(
            str(scenario.solver.directory),
            configuration_ids,
            [{}] * len(configuration_ids),
        )
        data_target.save_csv()
        # Submit the configuration job
        runs = [
            rrr.add_to_queue(
                runner=run_on,
                cmd=configuration_commands,
                name=f"{self.name}: {scenario.solver.name} on {scenario.instance_set.name}",
                base_dir=base_dir,
                output_path=output,
                parallel_jobs=num_parallel_jobs,
                sbatch_options=sbatch_options,
                prepend=slurm_prepend,
            )
        ]

        if validate_after:
            validate = scenario.solver.run_performance_dataframe(
                scenario.instance_set,
                config_ids=configuration_ids,
                performance_dataframe=data_target,
                cutoff_time=scenario.solver_cutoff_time,
                sbatch_options=sbatch_options,
                slurm_prepend=slurm_prepend,
                log_dir=scenario.validation,
                base_dir=base_dir,
                dependencies=runs,
                job_name=f"{self.name}: Validating {len(configuration_ids)} "
                f"{scenario.solver.name} Configurations on "
                f"{scenario.instance_set.name}",
                run_on=run_on,
            )
            runs.append(validate)

        if run_on == Runner.LOCAL:
            print(f"[{self.name}] Running {len(runs)} jobs locally...")
            for run in runs:
                run.wait()
            print(f"[{self.name}] Finished running {len(runs)} jobs locally.")
        return runs

    @staticmethod
    def organise_output(
        output_source: Path,
        output_target: Path,
        scenario: ConfigurationScenario,
        configuration_id: str,
    ) -> None | str:
        """Method to restructure and clean up after a single configurator call.

        Args:
            output_source: Path to the output file of the configurator run.
            output_target: Path to the Performance DataFrame to store result.
            scenario: ConfigurationScenario of the configuration.
            configuration_id: ID (of the run) of the configuration.
        """
        raise NotImplementedError

    @staticmethod
    def save_configuration(
        scenario: ConfigurationScenario,
        configuration_id: str,
        configuration: dict,
        output_target: Path,
    ) -> dict | None:
        """Method to save a configuration to a file.

        If the output_target is None, return the configuration.

        Args:
            scenario: ConfigurationScenario of the configuration. Should be removed.
            configuration_id: ID (of the run) of the configuration.
            configuration: Configuration to save.
            output_target: Path to the Performance DataFrame to store result.
        """
        if output_target is None or not output_target.exists():
            return configuration
        # Save result to Performance DataFrame
        from filelock import FileLock

        lock = FileLock(f"{output_target}.lock")
        with lock.acquire(timeout=600):
            performance_data = PerformanceDataFrame(output_target)
            # Resolve absolute path to Solver column
            solver = [
                s
                for s in performance_data.solvers
                if Path(s).name == scenario.solver.name
            ][0]
            # Update the configuration ID by adding the configuration
            performance_data.add_configuration(
                solver=solver,
                configuration_id=configuration_id,
                configuration=configuration,
            )
            performance_data.save_csv()

    def get_status_from_logs(self: Configurator) -> None:
        """Method to scan the log files of the configurator for warnings."""
        raise NotImplementedError


class ConfigurationScenario:
    """Template class to handle a configuration scenarios."""

    def __init__(
        self: ConfigurationScenario,
        solver: Solver,
        instance_set: InstanceSet,
        sparkle_objectives: list[SparkleObjective],
        number_of_runs: int,
        parent_directory: Path,
        timestamp: str = None,
    ) -> None:
        """Initialize scenario paths and names.

        Args:
            solver: Solver that should be configured.
            instance_set: Instances object for the scenario.
            sparkle_objectives: Sparkle Objectives to optimize.
            number_of_runs: The number of configurator runs to perform.
            parent_directory: Directory in which the scenario should be placed.
            timestamp: The timestamp of the scenario directory/file creation.
                Only set when read from file, otherwise generated at time of creation.
        """
        self.solver = solver
        self.instance_set = instance_set
        self.sparkle_objectives = sparkle_objectives
        self.number_of_runs = number_of_runs
        self.parent_directory = parent_directory
        self._timestamp = timestamp
        self._ablation_scenario: AblationScenario = None

    @property
    def configurator(self: ConfigurationScenario) -> Configurator:
        """Return the type of configurator the scenario belongs to."""
        return Configurator

    @property
    def name(self: ConfigurationScenario) -> str:
        """Return the name of the scenario."""
        return f"{self.solver.name}_{self.instance_set.name}_{self.timestamp}"

    @property
    def timestamp(self: ConfigurationScenario) -> str:
        """Return the timestamp."""
        return self._timestamp

    @property
    def directory(self: ConfigurationScenario) -> Path:
        """Return the path of the scenario directory."""
        return None if self.timestamp is None else self.parent_directory / self.name

    @property
    def scenario_file_path(self: ConfigurationScenario) -> Path:
        """Return the path of the scenario file."""
        if self.directory:
            return self.directory / "scenario.txt"
        return None

    @property
    def validation(self: ConfigurationScenario) -> Path:
        """Return the path of the validation directory."""
        if self.directory:
            return self.directory / "validation"
        return None

    @property
    def tmp(self: ConfigurationScenario) -> Path:
        """Return the path of the tmp directory."""
        if self.directory:
            return self.directory / "tmp"
        return None

    @property
    def results_directory(self: ConfigurationScenario) -> Path:
        """Return the path of the results directory."""
        if self.directory:
            return self.directory / "results"
        return None

    @property
    def configuration_ids(self: ConfigurationScenario) -> list[str]:
        """Return the IDs of the configurations for the scenario.

        Only exists after the scenario has been created.

        Returns:
            List of configuration IDs, one for each run.
        """
        return [
            f"{self.configurator.__name__}_{self.timestamp}_{i}"
            for i in range(self.number_of_runs)
        ]

    @property
    def ablation_scenario(self: ConfigurationScenario) -> AblationScenario:
        """Return the ablation scenario for the scenario if it exists."""
        if self._ablation_scenario is not None:
            return self._ablation_scenario
        for scenario in self.directory.glob("*/ablation_config.txt"):
            self._ablation_scenario = AblationScenario.from_file(scenario, self)
            return self._ablation_scenario
        return None

    def create_scenario(self: ConfigurationScenario) -> None:
        """Create scenario with solver and instances in the parent directory.

        This prepares all the necessary subdirectories related to configuration.

        Args:
            parent_directory: Directory in which the scenario should be created.
        """
        self._timestamp = datetime.now().strftime("%Y%m%d-%H%M")
        # Prepare scenario directory
        shutil.rmtree(self.directory, ignore_errors=True)
        self.directory.mkdir(parents=True)
        # Create empty directories as needed
        self.tmp.mkdir(exist_ok=True)
        self.validation.mkdir(exist_ok=True)
        self.results_directory.mkdir(exist_ok=True)

    def create_scenario_file(self: ConfigurationScenario) -> Path:
        """Create a file with the configuration scenario."""
        raise NotImplementedError

    def serialise(self: ConfigurationScenario) -> dict:
        """Serialize the configuration scenario."""
        raise NotImplementedError

    @classmethod
    def find_scenario(
        cls: ConfigurationScenario,
        directory: Path,
        solver: Solver,
        instance_set: InstanceSet,
        timestamp: str = None,
    ) -> ConfigurationScenario:
        """Resolve a scenario from a directory and Solver / Training set."""
        if timestamp is None:
            # Get the newest timestamp
            timestamp_list: list[datetime] = []
            for subdir in directory.iterdir():
                if subdir.is_dir():
                    dir_timestamp = subdir.name.split("_")[-1]
                    try:
                        dir_timestamp = datetime.strptime(dir_timestamp, "%Y%m%d-%H%M")
                        timestamp_list.append(dir_timestamp)
                    except ValueError:
                        continue

            if timestamp_list == []:
                return None
            timestamp = max(timestamp_list).strftime("%Y%m%d-%H%M")

        scenario_name = f"{solver.name}_{instance_set.name}_{timestamp}"
        path = directory / f"{scenario_name}" / "scenario.txt"
        if not path.exists():
            return None
        return cls.from_file(path)

    @staticmethod
    def from_file(scenario_file: Path) -> ConfigurationScenario:
        """Reads scenario file and initalises ConfigurationScenario."""
        raise NotImplementedError


class AblationScenario:
    """Class for ablation analysis."""

    # We use the SMAC2 target algorithm for solver output handling
    configurator_target = (
        Path(__file__).parent.resolve()
        / "implementations"
        / "SMAC2"
        / "smac2_target_algorithm.py"
    )

    ablation_dir = Path(__file__).parent / "implementations" / "ablationAnalysis-0.9.4"
    ablation_executable = ablation_dir / "ablationAnalysis"
    ablation_validation_executable = ablation_dir / "ablationValidation"

    def __init__(
        self: AblationScenario,
        configuration_scenario: ConfigurationScenario,
        test_set: InstanceSet,
        cutoff_length: str,
        concurrent_clis: int,
        best_configuration: dict,
        ablation_racing: bool = False,
    ) -> None:
        """Initialize ablation scenario.

        Args:
            solver: Solver object
            configuration_scenario: Configuration scenario
            train_set: The training instance
            test_set: The test instance
            cutoff_length: The cutoff length for ablation analysis
            concurrent_clis: The maximum number of concurrent jobs on a single node
            best_configuration: The configuration to ablate from.
            ablation_racing: Whether to use ablation racing
        """
        self.config_scenario = configuration_scenario
        self.solver = configuration_scenario.solver
        self.train_set = configuration_scenario.instance_set
        self.test_set = test_set
        self.cutoff_time = configuration_scenario.solver_cutoff_time
        self.cutoff_length = cutoff_length
        self.concurrent_clis = concurrent_clis
        self.best_configuration = best_configuration
        self.ablation_racing = ablation_racing
        self.scenario_name = f"ablation_{configuration_scenario.name}"
        self._table_file: Optional[Path] = None
        if self.test_set is not None:
            self.scenario_name += f"_{self.test_set.name}"

    @property
    def scenario_dir(self: AblationScenario) -> Path:
        """Return the path of the scenario directory."""
        if self.config_scenario.directory:
            return self.config_scenario.directory / self.scenario_name
        return None

    @property
    def tmp_dir(self: AblationScenario) -> Path:
        """Return the path of the tmp directory."""
        if self.scenario_dir:
            return self.scenario_dir / "tmp"
        return None

    @property
    def validation_dir(self: AblationScenario) -> Path:
        """Return the path of the validation directory."""
        if self.scenario_dir:
            return self.scenario_dir / "validation"
        return None

    @property
    def validation_dir_tmp(self: AblationScenario) -> Path:
        """Return the path of the validation tmp directory."""
        if self.validation_dir:
            return self.validation_dir / "tmp"
        return None

    @property
    def table_file(self: AblationScenario) -> Path:
        """Return the path of the table file."""
        if self._table_file:
            return self._table_file
        elif self.validation_dir:
            return self.validation_dir / "log" / "ablation-validation-run1234.txt"
        else:
            return None

    @staticmethod
    def check_requirements(verbose: bool = False) -> bool:
        """Check if Ablation Analysis is installed."""
        import warnings

        if no_java := shutil.which("java") is None:
            if verbose:
                warnings.warn(
                    "AblationAnalysis requires Java 1.8.0_402, but Java is not installed"
                    ". Please ensure Java is installed."
                )
        if no_exec := not AblationScenario.ablation_executable.exists():
            if verbose:
                warnings.warn(
                    "AblationAnalysis executable not found. Please ensure Ablation"
                    " Analysis is installed in the expected Path "
                    f"({AblationScenario.ablation_executable})."
                )
        if no_validation := not AblationScenario.ablation_validation_executable.exists():
            if verbose:
                warnings.warn(
                    "AblationAnalysis Validation executable not found. Please ensure "
                    "Ablation Analysis is installed in the expected Path "
                    f"({AblationScenario.ablation_validation_executable})."
                )
        return not (no_java or no_exec or no_validation)

    @staticmethod
    def download_requirements(
        ablation_url: str = "https://github.com/ADA-research/Sparkle/raw/refs/heads/development"
        "/Resources/Other/ablationAnalysis-0.9.4.zip",
    ) -> None:
        """Download Ablation Analysis executable."""
        if AblationScenario.ablation_executable.exists():
            return  # Already installed
        from urllib.request import urlopen
        import zipfile
        import io

        AblationScenario.ablation_dir.mkdir(parents=True, exist_ok=True)
        r = urlopen(ablation_url, timeout=60)
        z = zipfile.ZipFile(io.BytesIO(r.read()))
        z.extractall(AblationScenario.ablation_dir)
        # Ensure execution rights
        AblationScenario.ablation_executable.chmod(0o755)
        AblationScenario.ablation_validation_executable.chmod(0o755)

    def create_configuration_file(self: AblationScenario) -> Path:
        """Create a configuration file for ablation analysis.

        Returns:
            Path to the created configuration file.
        """
        objective = self.config_scenario.sparkle_objectives[0]
        pcs = self.solver.get_configuration_space()
        parameter_names = [p.name for p in pcs.values()]
        # We need to remove any redundant keys that are not in PCS
        best_configuration = self.best_configuration.copy()
        removable_keys = [
            key for key in best_configuration if key not in parameter_names
        ]
        for key in removable_keys:
            del best_configuration[key]
        opt_config_str = " ".join([f"-{k} {v}" for k, v in best_configuration.items()])
        # We need to check which params are missing and supplement with default values
        for p in list(pcs.values()):
            if p.name not in opt_config_str:
                opt_config_str += f" -{p.name} {p.default_value}"

        # Ablation cannot deal with E scientific notation in floats
        ctx = decimal.Context(prec=16)
        for config in opt_config_str.split(" -"):
            _, value = config.strip().split(" ")
            if "e" in value.lower():
                value = value.strip("'")
                float_value = float(value.lower())
                formatted = format(ctx.create_decimal(float_value), "f")
                opt_config_str = opt_config_str.replace(value, formatted)

        smac_run_obj = "RUNTIME" if objective.time else "QUALITY"
        objective_str = "MEAN10" if objective.time else "MEAN"
        pcs_file_path = f"{self.config_scenario.solver.pcs_file.absolute()}"

        # Create config file
        config_file = self.scenario_dir / "ablation_config.txt"
        config = (
            f'algo = "{AblationScenario.configurator_target.absolute()} '
            f"{self.config_scenario.solver.directory.absolute()} "
            f'{self.tmp_dir.absolute()} {objective}"\n'
            f"execdir = {self.tmp_dir.absolute()}\n"
            "experimentDir = ./\n"
            f"deterministic = {1 if self.solver.deterministic else 0}\n"
            f"run_obj = {smac_run_obj}\n"
            f"overall_obj = {objective_str}\n"
            f"cutoffTime = {self.cutoff_time}\n"
            f"cutoff_length = {self.cutoff_length}\n"
            f"cli-cores = {self.concurrent_clis}\n"
            f"useRacing = {self.ablation_racing}\n"
            f"seed = {random.randint(0, 2**32 - 1)}\n"
            f"paramfile = {pcs_file_path}\n"
            "instance_file = instances_train.txt\n"
            "test_instance_file = instances_test.txt\n"
            "sourceConfiguration = DEFAULT\n"
            f'targetConfiguration = "{opt_config_str}"'
        )
        config_file.open("w").write(config)
        # Write config to validation directory
        conf_valid = config.replace(
            f"execdir = {self.tmp_dir.absolute()}\n",
            f"execdir = {self.validation_dir_tmp.absolute()}\n",
        )
        (self.validation_dir / config_file.name).open("w").write(conf_valid)
        return self.validation_dir / config_file.name

    def create_instance_file(self: AblationScenario, test: bool = False) -> Path:
        """Create an instance file for ablation analysis."""
        file_suffix = "_train.txt"
        instance_set = self.train_set
        if test:
            file_suffix = "_test.txt"
            instance_set = self.test_set if self.test_set is not None else self.train_set
        # We give the Ablation script the paths of the instances
        file_instance = self.scenario_dir / f"instances{file_suffix}"
        with file_instance.open("w") as fh:
            for instance in instance_set._instance_paths:
                # We need to unpack the multi instance file paths in quotes
                if isinstance(instance, list):
                    joined_instances = " ".join(
                        [str(file.absolute()) for file in instance]
                    )
                    fh.write(f"{joined_instances}\n")
                else:
                    fh.write(f"{instance.absolute()}\n")
        # Copy to validation directory
        shutil.copyfile(file_instance, self.validation_dir / file_instance.name)
        return file_instance

    def create_scenario(self: AblationScenario, override_dirs: bool = False) -> None:
        """Create scenario directory and files."""
        if self.scenario_dir.exists():
            print("WARNING: Found existing ablation scenario.")
            if not override_dirs:
                print("Set override to True to overwrite existing scenario.")
                return
            print("Overwriting existing scenario...")
            shutil.rmtree(self.scenario_dir)
        self.tmp_dir.mkdir(parents=True, exist_ok=True)
        self.validation_dir_tmp.mkdir(parents=True, exist_ok=True)
        self.create_instance_file()
        self.create_instance_file(test=True)
        self.create_configuration_file()

    def check_for_ablation(self: AblationScenario) -> bool:
        """Checks if ablation has terminated successfully."""
        if not self.table_file.is_file():
            return False
        # First line in the table file should be "Ablation analysis validation complete."
        table_line = self.table_file.open().readline().strip()
        return table_line == "Ablation analysis validation complete."

    def read_ablation_table(self: AblationScenario) -> list[list[str]]:
        """Read from ablation table of a scenario."""
        if not self.check_for_ablation():
            # No ablation table exists for this solver-instance pair
            return []
        results = [
            [
                "Round",
                "Flipped parameter",
                "Source value",
                "Target value",
                "Validation result",
            ]
        ]

        for line in self.table_file.open().readlines():
            # Pre-process lines from the ablation file and add to the results dictionary.
            # Sometimes ablation rounds switch multiple parameters at once.
            # EXAMPLE: 2 EDR, EDRalpha   0, 0.1   1, 0.1013241633106732 486.31691
            # To split the row correctly, we remove the space before the comma separated
            # parameters and add it back.
            # T.S. 30-01-2024: the results object is a nested list not dictionary?
            values = re.sub(r"\s+", " ", line.strip())
            values = re.sub(r", ", ",", values)
            values = [val.replace(",", ", ") for val in values.split(" ")]
            if len(values) == 5:
                results.append(values)
        return results

    def submit_ablation(
        self: AblationScenario,
        log_dir: Path,
        sbatch_options: list[str] = [],
        slurm_prepend: str | list[str] | Path = None,
        run_on: Runner = Runner.SLURM,
    ) -> list[Run]:
        """Submit an ablation job.

        Args:
            log_dir: Directory to store job logs
            sbatch_options: Options to pass to sbatch
            slurm_prepend: Script to prepend to sbatch script
            run_on: Determines to which RunRunner queue the job is added

        Returns:
            A  list of Run objects. Empty when running locally.
        """
        if not self.check_requirements(verbose=True):
            raise RuntimeError(
                "Ablation Analysis is not available. Please ensure Java and Ablation "
                "Analysis is installed and try again."
            )
        # 1. submit the ablation to the runrunner queue
        cmd = (
            f"{AblationScenario.ablation_executable.absolute()} "
            "--optionFile ablation_config.txt"
        )
        srun_options = ["-N1", "-n1", f"-c{self.concurrent_clis}"]
        sbatch_options += [f"--cpus-per-task={self.concurrent_clis}"]
        run_ablation = rrr.add_to_queue(
            runner=run_on,
            cmd=cmd,
            name=f"Ablation analysis: {self.solver.name} on {self.train_set.name}",
            base_dir=log_dir,
            path=self.scenario_dir,
            sbatch_options=sbatch_options,
            srun_options=srun_options,
            prepend=slurm_prepend,
        )

        runs = []
        if run_on == Runner.LOCAL:
            run_ablation.wait()
        runs.append(run_ablation)

        # 2. Run ablation validation run if we have a test set to run on
        if self.test_set is not None:
            # Validation dir should have a copy of all needed files, except for the
            # output of the ablation run, which is stored in ablation-run[seed].txt
            cmd = (
                f"{AblationScenario.ablation_validation_executable.absolute()} "
                "--optionFile ablation_config.txt "
                "--ablationLogFile ../log/ablation-run1234.txt"
            )

            run_ablation_validation = rrr.add_to_queue(
                runner=run_on,
                cmd=cmd,
                name=f"Ablation validation: Test set {self.test_set.name}",
                path=self.validation_dir,
                base_dir=log_dir,
                dependencies=run_ablation,
                sbatch_options=sbatch_options,
                prepend=slurm_prepend,
            )

            if run_on == Runner.LOCAL:
                run_ablation_validation.wait()
            runs.append(run_ablation_validation)
        return runs

    @staticmethod
    def from_file(
        path: Path, config_scenario: ConfigurationScenario
    ) -> AblationScenario:
        """Reads scenario file and initalises AblationScenario."""
        variables = {}
        for line in path.open().readlines():
            if line.strip() == "":
                continue
            key, value = line.strip().split(" = ", maxsplit=1)
            variables[key] = value
        best_conf = {}
        for keyvalue in variables["targetConfiguration"].replace('"', "").split("-"):
            keyvalue = keyvalue.strip()
            if keyvalue:
                key, value = keyvalue.strip().split(" ", maxsplit=1)
                best_conf[key] = value
        test_set = None
        if (path.parent / "instances_test.txt").exists():
            test_path = (path.parent / "instances_test.txt").open().readline().strip()
            test_path = Path(test_path).parent
            if test_path != config_scenario.instance_set.directory:
                test_set = Instance_Set(test_path)
        return AblationScenario(
            config_scenario,
            test_set,
            variables["cutoff_length"],
            int(variables["cli-cores"]),
            best_conf,
            ablation_racing=bool(variables["useRacing"]),
        )



"""Resolver class if RunSolver is not supported."""

from pathlib import Path

from sparkle.tools.runsolver.runsolver import RunSolver
from sparkle.tools.runsolver.py_runsolver import PyRunSolver


class RunSolverResolver(RunSolver):
    """Class representation the RunSolver Resolver."""

    @staticmethod
    def wrap_command(
        runsolver_executable: Path,
        command: list[str],
        cutoff_time: int,
        log_directory: Path,
        log_name_base: str = None,
        raw_results_file: bool = True,
    ) -> list[str]:
        """Resolves if RunSolver is available. Otherwise uses a python implementation.

        Args:
            runsolver_executable: The Path to the runsolver executable.
                Is returned as an *absolute* path in the output.
            command: The command to wrap.
            cutoff_time: The cutoff CPU time for the solver.
            log_directory: The directory where to write the solver output.
            log_name_base: A user defined name to easily identify the logs.
                Defaults to "runsolver".
            raw_results_file: Whether to use the raw results file.

        Returns:
            List of commands and arguments to execute the solver.
        """
        if runsolver_executable.exists():
            return RunSolver.wrap_command(
                runsolver_executable,
                command,
                cutoff_time,
                log_directory,
                log_name_base,
                raw_results_file,
            )
        else:
            return PyRunSolver.wrap_command(
                command, cutoff_time, log_directory, log_name_base, raw_results_file
            )



#!/usr/bin/env python3
"""Command to remove temporary files not affecting the platform state."""

import re
import sys
import argparse
import shutil

from sparkle.structures import PerformanceDataFrame

from sparkle.CLI.help import logging as sl
from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help import snapshot_help as snh
from sparkle.CLI.help import jobs as jobs_help


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Command to clean files from the platform."
    )
    parser.add_argument(*ac.CleanupArgumentAll.names, **ac.CleanupArgumentAll.kwargs)
    parser.add_argument(*ac.CleanupArgumentLogs.names, **ac.CleanupArgumentLogs.kwargs)
    parser.add_argument(
        *ac.CleanupArgumentRemove.names, **ac.CleanupArgumentRemove.kwargs
    )
    parser.add_argument(
        *ac.CleanUpPerformanceDataArgument.names,
        **ac.CleanUpPerformanceDataArgument.kwargs,
    )
    return parser


def check_logs_performance_data(performance_data: PerformanceDataFrame) -> int:
    """Check if the performance data is missing values that can be extracted from the logs.

    Args:
        performance_data (PerformanceDataFrame): The performance data.

    Returns:
        int: The number of updated values.
    """
    # empty_indices = performance_data.empty_indices
    pattern = re.compile(
        r"^(?P<objective>\S+)\s*,\s*"
        r"(?P<instance>\S+)\s*,\s*"
        r"(?P<run_id>\S+)\s*\|\s*"
        r"(?P<solver>\S+)\s*,\s*"
        r"(?P<config_id>\S+)\s*:\s*"
        r"(?P<target_value>\S+)$"
    )
    import math

    # Only iterate over slurm log files
    log_files = [
        f
        for f in gv.settings().DEFAULT_log_output.glob("**/*")
        if f.is_file() and f.suffix == ".out"
    ]
    count = 0
    for log in log_files:
        for line in log.read_text().splitlines():
            match = pattern.match(line)
            if match:
                objective = match.group("objective")
                instance = match.group("instance")
                run_id = int(match.group("run_id"))
                solver = match.group("solver")
                config_id = match.group("config_id")
                target_value = match.group("target_value")
                current_value = performance_data.get_value(
                    solver, instance, config_id, objective, run_id
                )
                # TODO: Would be better to extract all nan indices from PDF and check against this?
                if (
                    (
                        isinstance(current_value, (int, float))
                        and math.isnan(current_value)
                    )
                    or isinstance(current_value, str)
                    and current_value == "nan"
                ):
                    performance_data.set_value(
                        target_value, solver, instance, config_id, objective, run_id
                    )
                    count += 1
    if count:
        performance_data.save_csv()
    return count


def remove_temporary_files() -> None:
    """Remove temporary files. Only removes files not affecting the sparkle state."""
    shutil.rmtree(gv.settings().DEFAULT_log_output, ignore_errors=True)
    gv.settings().DEFAULT_log_output.mkdir()


def main(argv: list[str]) -> None:
    """Main function of the cleanup command."""
    # Log command call
    sl.log_command(sys.argv, gv.settings().random_state)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)

    if args.performance_data:
        # Check if we can cleanup the PerformanceDataFrame if necessary
        from runrunner.base import Status

        running_jobs = jobs_help.get_runs_from_file(
            gv.settings().DEFAULT_log_output, filter=[Status.WAITING, Status.RUNNING]
        )
        if len(running_jobs) > 0:
            print("WARNING: There are still running jobs! Continue cleaning? [y/n]")
            a = input()
            if a != "y":
                sys.exit(0)

        performance_data = PerformanceDataFrame(
            gv.settings().DEFAULT_performance_data_path
        )
        count = check_logs_performance_data(performance_data)
        print(
            f"Extracted {count} values from the logs and placed them in the PerformanceDataFrame."
        )

        # Remove empty configurations
        removed_configurations = 0
        for solver, configurations in performance_data.configurations.items():
            for config_id, config in configurations.items():
                if config_id == PerformanceDataFrame.default_configuration:
                    continue
                if not config:  # Empty configuration, remove
                    performance_data.remove_configuration(solver, config_id)
                    removed_configurations += 1
        if removed_configurations:
            performance_data.save_csv()
        print(
            f"Removed {removed_configurations} empty configurations from the "
            "Performance DataFrame."
        )

        index_num = len(performance_data.index)
        # We only clean lines that are completely empty
        performance_data.remove_empty_runs()
        performance_data.save_csv()
        print(
            f"Removed {index_num - len(performance_data.index)} rows from the "
            f"Performance DataFrame, leaving {len(performance_data.index)} rows."
        )

    if args.all:
        shutil.rmtree(gv.settings().DEFAULT_output, ignore_errors=True)
        snh.create_working_dirs()
        print("Removed all output files from the platform!")
    elif args.remove:
        snh.remove_current_platform()
        snh.create_working_dirs()
        print("Cleaned platform of all files!")
    elif args.logs:
        remove_temporary_files()
        print("Cleaned platform of log files!")
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
"""Sparkle command to remove an instance set from the Sparkle platform."""

import sys
import argparse
import shutil

from sparkle.CLI.help import global_variables as gv
from sparkle.platform import file_help as sfh
from sparkle.structures import FeatureDataFrame, PerformanceDataFrame
from sparkle.instance import Instance_Set
from sparkle.CLI.help import logging as sl
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.nicknames import resolve_object_name


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(description="Remove instances from the platform.")
    parser.add_argument(
        *ac.InstancesPathRemoveArgument.names, **ac.InstancesPathRemoveArgument.kwargs
    )
    return parser


def main(argv: list[str]) -> None:
    """Main function of the remove instances command."""
    # Log command call
    sl.log_command(sys.argv, gv.settings().random_state)
    check_for_initialise()

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)
    instances_path = resolve_object_name(
        args.instances_path,
        nickname_dict=gv.file_storage_data_mapping[gv.instances_nickname_path],
        target_dir=gv.settings().DEFAULT_instance_dir,
    )

    if (
        instances_path is None
        or not instances_path.exists()
        or not instances_path.is_dir()
    ):
        print(f'Could not resolve instances path arg "{args.instances_path}"!')
        print("Check that the path or nickname is spelled correctly.")
        sys.exit(-1)

    print(f"Start removing all instances in directory {instances_path} ...")
    old_instance_set = Instance_Set(instances_path)
    # Remove from feature data and performance data
    feature_data = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)
    performance_data = PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path)
    for instance in old_instance_set.instance_names:
        feature_data.remove_instances(instance)
        performance_data.remove_instances(instance)

    feature_data.save_csv()
    performance_data.save_csv()

    # Remove nickname, if it exists
    instances_nicknames = gv.file_storage_data_mapping[gv.instances_nickname_path]
    for key in instances_nicknames:
        if instances_nicknames[key] == instances_path:
            sfh.add_remove_platform_item(
                instances_path, gv.instances_nickname_path, key=key, remove=True
            )
            break

    # We unlink symbolics links, erase copies
    if instances_path.is_symlink():
        instances_path.unlink()
    else:
        # Remove the directory and all its files
        shutil.rmtree(instances_path)

    print(f"Removing instances set {instances_path.name} done!")
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
"""Sparkle command to show information about Sparkle."""

import sys
import argparse
import sparkle.__about__


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments.

    Returns:
      The argument parser.
    """
    parser = argparse.ArgumentParser("Show information about Sparkle.")
    return parser


def main(argv: list[str]) -> None:
    """Main function of the command."""
    authors = sparkle.__about__.authors.split(", ")
    print(
        "\n".join(
            [
                f"Sparkle ({sparkle.__about__.description})",
                f"Version: {sparkle.__about__.__version__}",
                f"Licence: {sparkle.__about__.licence}",
                f"Written by {', '.join(authors[:-1])} and {authors[-1]}",
                f"Contact: {sparkle.__about__.contact}",
                "For more details see README.md",
            ]
        )
    )
    sys.exit()


if __name__ == "__main__":
    main(sys.argv[1:])



"""Class for solver status."""

from __future__ import annotations
from enum import Enum


class SolverStatus(str, Enum):
    """Possible return states for solver runs."""

    SUCCESS = "SUCCESS"  # Positive status
    SAT = "SAT"  # SAT specific positive status
    UNSAT = "UNSAT"  # SAT specific positive status

    # Negative status
    UNKNOWN = "UNKNOWN"
    CRASHED = "CRASHED"
    TIMEOUT = "TIMEOUT"
    WRONG = "WRONG"
    ERROR = "ERROR"
    KILLED = "KILLED"

    def __str__(self: SolverStatus) -> str:
        """Return the string value of the SolverStatus."""
        return str(self.value)

    @property
    def positive(self: SolverStatus) -> bool:
        """Return whether the status is positive."""
        return self in [SolverStatus.SUCCESS, SolverStatus.SAT, SolverStatus.UNSAT]



"""This module provides tools for the argument parsing for solver wrappers."""

from pathlib import Path
import ast
from typing import Any

from sparkle.types import resolve_objective


def parse_commandline_dict(args: list[str]) -> dict:
    """Parses a commandline dictionary to the object."""
    dict_str = " ".join(args)
    dict_str = dict_str[
        dict_str.index("{") : dict_str.index("}") + 1
    ]  # Slurm script fix
    return ast.literal_eval(dict_str)


def parse_instance(instance_str: str) -> Path | list[Path]:
    """Handles the parsing of multi file instances."""
    instance_str = instance_str.removeprefix("[")
    instance_str = instance_str.removesuffix("]")
    instance_list = instance_str.split(",")
    instance_list = [inst_str.strip(' "') for inst_str in instance_list]
    return Path(instance_list[0]) if len(instance_list) == 1 else instance_list


def parse_solver_wrapper_args(args: list[str]) -> dict[Any]:
    """Parse the arguments passed to the solver wrapper.

    Args:
        args: a list of arguments passed via the command line. It is ensured by Sparkle
            that this list contains certain keys such as `solver_dir`.

    Returns:
        A dictionary mapping argument names to their currently held values.
    """
    args_dict = parse_commandline_dict(args)

    # Some data needs specific formatting
    args_dict["solver_dir"] = Path(args_dict["solver_dir"])
    instance = args_dict["instance"]
    args_dict["instance"] = parse_instance(instance)
    args_dict["seed"] = int(args_dict["seed"])
    args_dict["objectives"] = [
        resolve_objective(name) for name in args_dict["objectives"].split(",")
    ]
    args_dict["cutoff_time"] = float(args_dict["cutoff_time"])
    return args_dict


def get_solver_call_params(
    args_dict: dict, prefix: str = "-", postfix: str = " "
) -> list[str]:
    """Gather the additional parameters for the solver call.

    Args:
        args_dict: Dictionary mapping argument names to their currently held values
        prefix: Prefix of the command line options
        postfix: Postfix of the command line options

    Returns:
        A list of parameters for the solver call
    """
    params = []
    # Certain arguments are not relevant/have already been processed
    ignore_args = {"solver_dir", "instance", "cutoff_time", "seed", "objectives"}
    for key in args_dict:
        if key not in ignore_args and args_dict[key] is not None:
            if postfix == " ":
                params.extend([prefix + str(key), str(args_dict[key])])
            else:
                params.extend([prefix + str(key) + postfix + str(args_dict[key])])

    return params



"""Module to manage performance data files and common operations on them."""

from __future__ import annotations
import ast
import copy
from typing import Any
import itertools
from pathlib import Path
import math
import numpy as np
import pandas as pd

from sparkle.types import SparkleObjective, resolve_objective


class PerformanceDataFrame(pd.DataFrame):
    """Class to manage performance data and common operations on them."""

    missing_value = math.nan

    missing_objective = "UNKNOWN"
    default_configuration = "Default"

    index_objective = "Objective"
    index_instance = "Instance"
    index_run = "Run"
    multi_index_names = [index_objective, index_instance, index_run]

    column_solver = "Solver"
    column_configuration = "Configuration"
    column_meta = "Meta"
    column_value = "Value"
    column_seed = "Seed"
    multi_column_names = [column_solver, column_configuration, column_meta]
    multi_column_value = [column_value, column_seed]
    multi_column_dtypes = [str, int]

    def __init__(
        self: PerformanceDataFrame,
        csv_filepath: Path,
        solvers: list[str] = None,
        configurations: dict[str, dict[str, dict]] = None,
        objectives: list[str | SparkleObjective] = None,
        instances: list[str] = None,
        n_runs: int = 1,
    ) -> None:
        """Initialise a PerformanceDataFrame.

        Consists of:
            - Columns representing the Solvers
            - Rows representing the result by multi-index in order of:
                * Objective (Static, given in constructor or read from file)
                * Instance
                * Runs (Static, given in constructor or read from file)

        Args:
            csv_filepath: If path exists, load from Path.
                Otherwise create new and save to this path.
            solvers: List of solver names to be added into the Dataframe
            configurations: The configuration keys per solver to add, structured as
                configurations[solver][config_key] = {"parameter": "value", ..}
            objectives: List of SparkleObjectives or objective names. By default None,
                then the objectives will be derived from Sparkle Settings if possible.
            instances: List of instance names to be added into the Dataframe
            n_runs: The number of runs to consider per Solver/Objective/Instance comb.
        """
        if csv_filepath and csv_filepath.exists():  # Read from file
            df = pd.read_csv(
                csv_filepath,
                header=[0, 1, 2],
                index_col=[0, 1, 2],
                on_bad_lines="skip",
                dtype={"Value": str, "Seed": int},
                comment="$",
            )  # $ For extra data lines
            super().__init__(df)
            self.csv_filepath = csv_filepath
            # Load configuration mapping
            with self.csv_filepath.open() as f:
                configuration_lines = [
                    line.strip().strip("$").split(",", maxsplit=2)
                    for line in f.readlines()
                    if line.startswith("$")
                ]
            configurations = {s: {} for s in self.solvers}
            for solver, config_key, config in configuration_lines[1:]:  # Skip header
                configurations[solver][config_key] = ast.literal_eval(config.strip('"'))
        else:  # New PerformanceDataFrame
            # Initialize empty DataFrame
            run_ids = list(range(1, n_runs + 1))  # We count runs from 1
            # We always need objectives to maintain the dimensions
            if objectives is None:
                objectives = [PerformanceDataFrame.missing_objective]
            else:
                objectives = [str(o) for o in objectives]
            # We always need an instance to maintain the dimensions
            if instances is None:
                instances = [PerformanceDataFrame.missing_value]
            # We always need a solver to maintain the dimensions
            if solvers is None:
                solvers = [PerformanceDataFrame.missing_value]
            midx = pd.MultiIndex.from_product(
                [objectives, instances, run_ids],
                names=PerformanceDataFrame.multi_index_names,
            )
            # Create the multi index tuples
            if configurations is None:
                configurations = {
                    solver: {PerformanceDataFrame.default_configuration: {}}
                    for solver in solvers
                }
            column_tuples = []
            # We cannot do .from_product here as config ids are per solver
            for solver in configurations.keys():
                for config_id in configurations[solver].keys():
                    column_tuples.extend(
                        [
                            (solver, config_id, PerformanceDataFrame.column_seed),
                            (solver, config_id, PerformanceDataFrame.column_value),
                        ]
                    )
            mcolumns = pd.MultiIndex.from_tuples(
                column_tuples,
                names=[
                    PerformanceDataFrame.column_solver,
                    PerformanceDataFrame.column_configuration,
                    PerformanceDataFrame.column_meta,
                ],
            )
            # Set dtype object to avoid inferring float for categorical objectives
            super().__init__(
                PerformanceDataFrame.missing_value,
                index=midx,
                columns=mcolumns,
                dtype="object",
            )
            self.csv_filepath = csv_filepath

        # Store configuration in global attributes dictionary, see Pandas Docs
        self.attrs = configurations

        if self.index.duplicated().any():  # Drop all duplicates except for last
            # NOTE: This is rather convoluted (but fast!) due to the fact we need to do it inplace to maintain our type (PerformanceDataFrame)
            # Make the index levels into columns (in-place)
            self.reset_index(inplace=True)
            # The first nlevels columns are the index columns created by reset_index, drop duplicates in those columns
            idx_cols = self.columns[
                : len(PerformanceDataFrame.multi_index_names)
            ].tolist()
            self.drop_duplicates(
                subset=idx_cols, keep="last", inplace=True
            )  # Drop duplicates
            self.set_index(idx_cols, inplace=True)  # Restore the MultiIndex (in-place)
            self.index.rename(
                self.multi_index_names, inplace=True
            )  # Restore level names

        # Sort the index to optimize lookup speed
        self.sort_index(axis=0, inplace=True)
        self.sort_index(axis=1, inplace=True)

        if csv_filepath and not self.csv_filepath.exists():  # New Performance DataFrame
            self.save_csv()

    # Properties

    @property
    def num_objectives(self: PerformanceDataFrame) -> int:
        """Retrieve the number of objectives in the DataFrame."""
        return self.index.get_level_values(0).unique().size

    @property
    def num_instances(self: PerformanceDataFrame) -> int:
        """Return the number of instances."""
        return self.index.get_level_values(1).unique().size

    @property
    def num_runs(self: PerformanceDataFrame) -> int:
        """Return the maximum number of runs of each instance."""
        return self.index.get_level_values(2).unique().size

    @property
    def num_solvers(self: PerformanceDataFrame) -> int:
        """Return the number of solvers."""
        return self.columns.get_level_values(0).unique().size

    @property
    def num_solver_configurations(self: PerformanceDataFrame) -> int:
        """Return the number of solver configurations."""
        return int(
            self.columns.get_level_values(  # Config has a seed & value
                PerformanceDataFrame.column_configuration
            ).size
            / 2
        )

    @property
    def multi_objective(self: PerformanceDataFrame) -> bool:
        """Return whether the dataframe represent MO or not."""
        return self.num_objectives > 1

    @property
    def solvers(self: PerformanceDataFrame) -> list[str]:
        """Return the solver present as a list of strings."""
        # Do not return the nan solver as its not an actual solver
        return (
            self.columns.get_level_values(PerformanceDataFrame.column_solver)
            .dropna()
            .unique()
            .to_list()
        )

    @property
    def configuration_ids(self: PerformanceDataFrame) -> list[str]:
        """Return the list of configuration keys."""
        return (
            self.columns.get_level_values(PerformanceDataFrame.column_configuration)
            .unique()
            .to_list()
        )

    @property
    def configurations(self: PerformanceDataFrame) -> dict[str, dict[str, dict]]:
        """Return a dictionary (copy) containing the configurations for each solver."""
        return copy.deepcopy(self.attrs)  # Deepcopy to avoid mutation of attribute

    @property
    def objective_names(self: PerformanceDataFrame) -> list[str]:
        """Return the objective names as a list of strings."""
        return self.index.get_level_values(0).unique().to_list()

    @property
    def objectives(self: PerformanceDataFrame) -> list[SparkleObjective]:
        """Return the objectives as a list of SparkleObjectives."""
        return [resolve_objective(o) for o in self.objective_names]

    @property
    def instances(self: PerformanceDataFrame) -> list[str]:
        """Return the instances as a Pandas Index object."""
        return self.index.get_level_values(1).unique().to_list()

    @property
    def run_ids(self: PerformanceDataFrame) -> list[int]:
        """Return the run ids as a list of integers."""
        return self.index.get_level_values(2).unique().to_list()

    @property
    def has_missing_values(self: PerformanceDataFrame) -> bool:
        """Returns True if there are any missing values in the dataframe."""
        return (
            self.drop(
                PerformanceDataFrame.column_seed,
                level=PerformanceDataFrame.column_meta,
                axis=1,
            )
            .isnull()
            .any()
            .any()
        )

    def is_missing(
        self: PerformanceDataFrame,
        solver: str,
        instance: str,
    ) -> int:
        """Checks if a solver/instance is missing values."""
        return (
            self.xs(solver, axis=1)
            .xs(instance, axis=0, level=PerformanceDataFrame.index_instance)
            .drop(
                PerformanceDataFrame.column_seed,
                level=PerformanceDataFrame.column_meta,
                axis=1,
            )
            .isnull()
            .any()
            .any()
        )

    def verify_objective(self: PerformanceDataFrame, objective: str) -> str:
        """Method to check whether the specified objective is valid.

        Users are allowed to index the dataframe without specifying all dimensions.
        However, when dealing with multiple objectives this is not allowed and this
        is verified here. If we have only one objective this is returned. Otherwise,
        if an objective is specified by the user this is returned.

        Args:
            objective: The objective given by the user
        """
        if objective is None:
            if self.multi_objective:
                raise ValueError("Error: MO Data, but objective not specified.")
            elif self.num_objectives == 1:
                return self.objective_names[0]
            else:
                return PerformanceDataFrame.missing_objective
        return objective

    def verify_run_id(self: PerformanceDataFrame, run_id: int) -> int:
        """Method to check whether run id is valid.

        Similar to verify_objective but here we check the dimensionality of runs.

        Args:
            run_id: the run as specified by the user.
        """
        if run_id is None:
            if self.num_runs > 1:
                raise ValueError(
                    "Error: Multiple run performance data, but run not specified"
                )
            else:
                run_id = self.run_ids[0]
        return run_id

    def verify_indexing(
        self: PerformanceDataFrame, objective: str, run_id: int
    ) -> tuple[str, int]:
        """Method to check whether data indexing is correct.

        Users are allowed to use the Performance Dataframe without the second and
        fourth dimension (Objective and Run respectively) in the case they only
        have one objective or only do one run. This method adjusts the indexing for
        those cases accordingly.

        Args:
            objective: The given objective name
            run_id: The given run index

        Returns:
            A tuple representing the (possibly adjusted) Objective and Run index.
        """
        objective = self.verify_objective(objective)
        run_id = self.verify_run_id(run_id)
        return objective, run_id

    # Getters and Setters

    def add_solver(
        self: PerformanceDataFrame,
        solver_name: str,
        configurations: list[(str, dict)] = None,
        initial_value: float | list[str | float] = None,
    ) -> None:
        """Add a new solver to the dataframe. Initializes value to None by default.

        Args:
            solver_name: The name of the solver to be added.
            configurations: A list of configuration keys for the solver.
            initial_value: The value assigned for each index of the new solver.
                If not None, must match the index dimension (n_obj * n_inst * n_runs).
        """
        if solver_name in self.solvers:
            print(
                f"WARNING: Tried adding already existing solver {solver_name} to "
                f"Performance DataFrame: {self.csv_filepath}"
            )
            return
        if not isinstance(initial_value, list):  # Single value
            initial_value = [[initial_value, initial_value]]
        if configurations is None:
            configurations = [(PerformanceDataFrame.default_configuration, {})]
        self.attrs[solver_name] = {}
        for (config_key, config), (value, seed) in itertools.product(
            configurations, initial_value
        ):
            self[(solver_name, config_key, PerformanceDataFrame.column_seed)] = seed
            self[(solver_name, config_key, PerformanceDataFrame.column_value)] = value
            self.attrs[solver_name][config_key] = config
        if self.num_solvers == 2:  # Remove nan solver
            for solver in self.solvers:
                if str(solver) == str(PerformanceDataFrame.missing_value):
                    self.remove_solver(solver)
                    break

    def add_configuration(
        self: PerformanceDataFrame,
        solver: str,
        configuration_id: str | list[str],
        configuration: dict[str, Any] | list[dict[str, Any]] = None,
    ) -> None:
        """Add new configurations for a solver to the dataframe.

        If the key already exists, update the value.

        Args:
            solver: The name of the solver to be added.
            configuration_id: The name of the configuration to be added.
            configuration: The configuration to be added.
        """
        if not isinstance(configuration_id, list):
            configuration_id = [configuration_id]
        if not isinstance(configuration, list):
            configuration = [configuration]
        for config_id, config in zip(configuration_id, configuration):
            if config_id not in self.get_configurations(solver):
                self[(solver, config_id, PerformanceDataFrame.column_value)] = None
                self[(solver, config_id, PerformanceDataFrame.column_seed)] = None
            self.attrs[solver][config_id] = config
        # Sort the index to optimize lookup speed
        self.sort_index(axis=1, inplace=True)

    def add_objective(
        self: PerformanceDataFrame, objective_name: str, initial_value: float = None
    ) -> None:
        """Add an objective to the DataFrame."""
        initial_value = initial_value or self.missing_value
        if objective_name in self.objective_names:
            print(
                f"WARNING: Tried adding already existing objective {objective_name} "
                f"to Performance DataFrame: {self.csv_filepath}"
            )
            return
        for instance, run in itertools.product(self.instances, self.run_ids):
            self.loc[(objective_name, instance, run)] = initial_value
        self.sort_index(axis=0, inplace=True)

    def add_instance(
        self: PerformanceDataFrame,
        instance_name: str,
        initial_values: Any | list[Any] = None,
    ) -> None:
        """Add and instance to the DataFrame.

        Args:
            instance_name: The name of the instance to be added.
            initial_values: The values assigned for each index of the new instance.
                If list, must match the column dimension (Value, Seed, Configuration).
        """
        initial_values = initial_values or self.missing_value
        if not isinstance(initial_values, list):
            initial_values = (
                [initial_values]
                * 2  # Value and Seed per target column
                * self.num_solver_configurations
            )
        elif len(initial_values) == len(PerformanceDataFrame.multi_column_names):
            initial_values = initial_values * self.num_solvers

        if instance_name in self.instances:
            print(
                f"WARNING: Tried adding already existing instance {instance_name} "
                f"to Performance DataFrame: {self.csv_filepath}"
            )
            return
        # Add rows for all combinations
        for objective, run in itertools.product(self.objective_names, self.run_ids):
            self.loc[(objective, instance_name, run)] = initial_values
        if self.num_instances == 2:  # Remove nan instance
            for instance in self.instances:
                if not isinstance(instance, str) and math.isnan(instance):
                    self.remove_instances(instance)
                    break
        # Sort the index to optimize lookup speed
        self.sort_index(axis=0, inplace=True)

    def add_runs(
        self: PerformanceDataFrame,
        num_extra_runs: int,
        instance_names: list[str] = None,
        initial_values: Any | list[Any] = None,
    ) -> None:
        """Add runs to the DataFrame.

        Args:
            num_extra_runs: The number of runs to be added.
            instance_names: The instances for which runs are to be added.
              By default None, which means runs are added to all instances.
            initial_values: The initial value for each objective of each new run.
                If a list, needs to have a value for Value, Seed and Configuration.
        """
        initial_values = initial_values or self.missing_value
        if not isinstance(initial_values, list):
            initial_values = [initial_values] * self.num_solvers * 2  # Value and Seed
        elif len(initial_values) == 2:  # Value and seed provided
            initial_values = initial_values * self.num_solvers
        instance_names = self.instances if instance_names is None else instance_names
        for objective, instance in itertools.product(
            self.objective_names, instance_names
        ):
            index_runs_start = len(self.loc[(objective, instance)]) + 1
            for run in range(index_runs_start, index_runs_start + num_extra_runs):
                self.loc[(objective, instance, run)] = initial_values
            # Sort the index to optimize lookup speed
            # NOTE: It would be better to do this at the end, but that results in
            # PerformanceWarning: indexing past lexsort depth may impact performance.
            self.sort_index(axis=0, inplace=True)

    def get_configurations(self: PerformanceDataFrame, solver_name: str) -> list[str]:
        """Return the list of configuration keys for a solver."""
        return list(
            self[solver_name]
            .columns.get_level_values(PerformanceDataFrame.column_configuration)
            .unique()
        )

    def get_full_configuration(
        self: PerformanceDataFrame, solver: str, configuration_id: str | list[str]
    ) -> dict | list[dict]:
        """Return the actual configuration associated with the configuration key."""
        if isinstance(configuration_id, str):
            return self.attrs[solver][configuration_id]
        return [self.attrs[solver][cid] for cid in configuration_id]

    def remove_solver(self: PerformanceDataFrame, solvers: str | list[str]) -> None:
        """Drop one or more solvers from the Dataframe."""
        if not solvers:  # Bugfix for when an empty list is passed to avoid nan adding
            return
        # To make sure objectives / runs are saved when no solvers are present
        solvers = [solvers] if isinstance(solvers, str) else solvers
        if self.num_solvers == 1:  # This would preferrably be done after removing
            for field in PerformanceDataFrame.multi_column_value:
                self[
                    PerformanceDataFrame.missing_value,
                    PerformanceDataFrame.missing_value,
                    field,
                ] = PerformanceDataFrame.missing_value
        self.drop(columns=solvers, level=0, axis=1, inplace=True)
        for solver in solvers:
            del self.attrs[solver]

    def remove_configuration(
        self: PerformanceDataFrame, solver: str, configuration: str | list[str]
    ) -> None:
        """Drop one or more configurations from the Dataframe."""
        if isinstance(configuration, str):
            configuration = [configuration]
        for config in configuration:
            self.drop((solver, config), axis=1, inplace=True)
            del self.attrs[solver][config]
        # Sort the index to optimize lookup speed
        self.sort_index(axis=1, inplace=True)

    def remove_objective(
        self: PerformanceDataFrame, objectives: str | list[str]
    ) -> None:
        """Remove objective from the Dataframe."""
        if len(self.objectives) < 2:
            raise Exception("Cannot remove last objective from PerformanceDataFrame")
        self.drop(
            objectives,
            axis=0,
            level=PerformanceDataFrame.index_objective,
            inplace=True,
        )

    def remove_instances(self: PerformanceDataFrame, instances: str | list[str]) -> None:
        """Drop instances from the Dataframe."""
        # To make sure objectives / runs are saved when no instances are present
        num_instances = len(instances) if isinstance(instances, list) else 1
        if self.num_instances - num_instances == 0:
            for objective, run in itertools.product(self.objective_names, self.run_ids):
                self.loc[(objective, PerformanceDataFrame.missing_value, run)] = (
                    PerformanceDataFrame.missing_value
                )
        self.drop(
            instances, axis=0, level=PerformanceDataFrame.index_instance, inplace=True
        )
        # Sort the index to optimize lookup speed
        self.sort_index(axis=0, inplace=True)

    def remove_runs(
        self: PerformanceDataFrame,
        runs: int | list[int],
        instance_names: list[str] = None,
    ) -> None:
        """Drop one or more runs from the Dataframe.

        Args:
            runs: The run indices to be removed. If its an int,
              the last n runs are removed. NOTE: If each instance has a different
              number of runs, the amount of removed runs is not uniform.
            instance_names: The instances for which runs are to be removed.
              By default None, which means runs are removed from all instances.
        """
        instance_names = self.instances if instance_names is None else instance_names
        runs = (
            list(range((self.num_runs + 1) - runs, (self.num_runs + 1)))
            if isinstance(runs, int)
            else runs
        )
        self.drop(runs, axis=0, level=PerformanceDataFrame.index_run, inplace=True)
        # Sort the index to optimize lookup speed
        self.sort_index(axis=0, inplace=True)

    def remove_empty_runs(self: PerformanceDataFrame) -> None:
        """Remove runs that contain no data, except for the first."""
        for row_index in self.index:
            if row_index[2] == 1:  # First run, never delete
                continue
            if self.loc[row_index].isna().all():
                self.drop(row_index, inplace=True)

    def filter_objective(self: PerformanceDataFrame, objective: str | list[str]) -> None:
        """Filter the Dataframe to a subset of objectives."""
        if isinstance(objective, str):
            objective = [objective]
        self.drop(
            list(set(self.objective_names) - set(objective)),
            axis=0,
            level=PerformanceDataFrame.index_objective,
            inplace=True,
        )

    def reset_value(
        self: PerformanceDataFrame,
        solver: str,
        instance: str,
        objective: str = None,
        run: int = None,
    ) -> None:
        """Reset a value in the dataframe."""
        self.set_value(
            PerformanceDataFrame.missing_value, solver, instance, objective, run
        )

    def set_value(
        self: PerformanceDataFrame,
        value: float | str | list[float | str] | list[list[float | str]],
        solver: str | list[str],
        instance: str | list[str],
        configuration: str = None,
        objective: str | list[str] = None,
        run: int | list[int] = None,
        solver_fields: list[str] = ["Value"],
        append_write_csv: bool = False,
    ) -> None:
        """Setter method to assign a value to the Dataframe.

        Allows for setting the same value to multiple indices.

        Args:
            value: Value(s) to be assigned. If value is a list, first dimension is
                the solver field, second dimension is if multiple different values are
                to be assigned. Must be the same shape as target.
            solver: The solver(s) for which the value should be set.
                If solver is a list, multiple solvers are set. If None, all
                solvers are set.
            instance: The instance(s) for which the value should be set.
                If instance is a list, multiple instances are set. If None, all
                instances are set.
            configuration: The configuration(s) for which the value should be set.
                When left None, set for all configurations
            objective: The objectives for which the value should be set.
                When left None, set for all objectives
            run: The run index for which the value should be set.
                If left None, set for all runs.
            solver_fields: The level to which each value should be assinged.
                Defaults to ["Value"].
            append_write_csv: For concurrent writing to the PerformanceDataFrame.
                If True, the value is directly appended to the CSV file.
                This will create duplicate entries in the file, but these are combined
                when loading the file.
        """
        # Convert indices to slices for None values
        solver = slice(solver) if solver is None else solver
        configuration = slice(configuration) if configuration is None else configuration
        instance = slice(instance) if instance is None else instance
        objective = slice(objective) if objective is None else objective
        run = slice(run) if run is None else run
        # Convert column indices to slices for setting multiple columns
        value = [value] if not isinstance(value, list) else value
        # NOTE: We currently forloop levels here, as it allows us to set the same
        # sequence of values to the indices
        for item, level in zip(value, solver_fields):
            self.loc[(objective, instance, run), (solver, configuration, level)] = item

        if append_write_csv:
            writeable = self.loc[(objective, instance, run), :]
            if isinstance(writeable, pd.Series):  # Single row, convert to pd.DataFrame
                writeable = self.loc[[(objective, instance, run)], :]
            # Append the new rows to the dataframe csv file
            writeable.to_csv(self.csv_filepath, mode="a", header=False)

    def get_value(
        self: PerformanceDataFrame,
        solver: str | list[str] = None,
        instance: str | list[str] = None,
        configuration: str = None,
        objective: str = None,
        run: int = None,
        solver_fields: list[str] = ["Value"],
    ) -> float | str | list[Any]:
        """Index a value of the DataFrame and return it."""
        # Convert indices to slices for None values
        solver = slice(solver) if solver is None else solver
        configuration = slice(configuration) if configuration is None else configuration
        instance = slice(instance) if instance is None else instance
        objective = slice(objective) if objective is None else objective
        solver_fields = slice(solver_fields) if solver_fields is None else solver_fields
        run = slice(run) if run is None else run
        target = self.loc[
            (objective, instance, run), (solver, configuration, solver_fields)
        ].values
        # Reduce dimensions when relevant
        if len(target) > 0 and isinstance(target[0], np.ndarray) and len(target[0]) == 1:
            target = target.flatten()
        target = target.tolist()
        if len(target) == 1:
            return target[0]
        return target

    def get_instance_num_runs(self: PerformanceDataFrame, instance: str) -> int:
        """Return the number of runs for an instance."""
        # We assume each objective has the same index for Instance/Runs
        return len(self.loc[(self.objective_names[0], instance)].index)

    # Calculables

    def mean(
        self: PerformanceDataFrame,
        objective: str = None,
        solver: str = None,
        instance: str = None,
    ) -> float:
        """Return the mean value of a slice of the dataframe."""
        objective = self.verify_objective(objective)
        subset = self.xs(objective, level=0)
        if solver is not None:
            subset = subset.xs(solver, axis=1, drop_level=False)
        if instance is not None:
            subset = subset.xs(instance, axis=0, drop_level=False)
        value = subset.astype(float).mean()
        if isinstance(value, pd.Series):
            return value.mean()
        return value

    def get_job_list(
        self: PerformanceDataFrame, rerun: bool = False
    ) -> list[tuple[str, str]]:
        """Return a list of performance computation jobs there are to be done.

        Get a list of tuple[instance, solver] to run from the performance data.
        If rerun is False (default), get only the tuples that don't have a
        value, else (True) get all the tuples.

        Args:
            rerun: Boolean indicating if we want to rerun all jobs

        Returns:
            A tuple of (solver, config, instance, run) combinations
        """
        # Drop the seed as we are looking for nan values, not seeds
        df = self.drop(
            PerformanceDataFrame.column_seed,
            axis=1,
            level=PerformanceDataFrame.column_meta,
        )
        df = df.droplevel(PerformanceDataFrame.column_meta, axis=1)
        if rerun:  # Return all combinations
            # Drop objective, not needed
            df = df.droplevel(PerformanceDataFrame.index_objective, axis=0)
            result = [
                tuple(column) + tuple(index)
                for column, index in itertools.product(df.columns, df.index)
            ]
        else:
            result = []
            for (solver, config), (objective, instance, run) in itertools.product(
                df.columns, df.index
            ):
                value = df.loc[(objective, instance, run), (solver, config)]
                if value is None or (
                    isinstance(value, (int, float)) and math.isnan(value)
                ):
                    result.append(tuple([solver, config, instance, run]))
        # Filter duplicates while keeping the order conistent
        result = list(dict.fromkeys(result))
        return result

    def configuration_performance(
        self: PerformanceDataFrame,
        solver: str,
        configuration: str | list[str] = None,
        objective: str | SparkleObjective = None,
        instances: list[str] = None,
        per_instance: bool = False,
    ) -> tuple[str, float]:
        """Return the (best) configuration performance for objective over the instances.

        Args:
            solver: The solver for which we determine evaluate the configuration
            configuration: The configuration (id) to evaluate
            objective: The objective for which we calculate find the best value
            instances: The instances which should be selected for the evaluation
            per_instance: Whether to return the performance per instance,
                or aggregated.

        Returns:
            The (best) configuration id and its aggregated performance.
        """
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        # Filter objective
        subdf = self.xs(objective.name, level=0, drop_level=True)
        # Filter solver
        subdf = subdf.xs(solver, axis=1, drop_level=True)
        # Drop the seed, then drop meta level as it is no longer needed
        subdf = subdf.drop(
            PerformanceDataFrame.column_seed,
            axis=1,
            level=PerformanceDataFrame.column_meta,
        )
        subdf = subdf.droplevel(PerformanceDataFrame.column_meta, axis=1)
        # Ensure the objective is numeric
        subdf = subdf.astype(float)

        if instances:  # Filter instances
            subdf = subdf.loc[instances, :]
        if configuration:  # Filter configuration
            if not isinstance(configuration, list):
                configuration = [configuration]
            subdf = subdf.filter(configuration, axis=1)
        # Aggregate the runs
        subdf = subdf.groupby(PerformanceDataFrame.index_instance).agg(
            func=objective.run_aggregator.__name__
        )
        # Aggregate the instances
        sub_series = subdf.agg(func=objective.instance_aggregator.__name__)
        # Select the best configuration
        best_conf = sub_series.idxmin() if objective.minimise else sub_series.idxmax()
        if per_instance:  # Return a list of instance results
            return best_conf, subdf[best_conf].to_list()
        return best_conf, sub_series[best_conf]

    def best_configuration(
        self: PerformanceDataFrame,
        solver: str,
        objective: SparkleObjective = None,
        instances: list[str] = None,
    ) -> tuple[str, float]:
        """Return the best configuration for the given objective over the instances.

        Args:
            solver: The solver for which we determine the best configuration
            objective: The objective for which we calculate the best configuration
            instances: The instances which should be selected for the evaluation

        Returns:
            The best configuration id and its aggregated performance.
        """
        return self.configuration_performance(solver, None, objective, instances)

    def best_instance_performance(
        self: PerformanceDataFrame,
        objective: str | SparkleObjective = None,
        instances: list[str] = None,
        run_id: int = None,
        exclude_solvers: list[(str, str)] = None,
    ) -> pd.Series:
        """Return the best performance for each instance in the portfolio.

        Args:
            objective: The objective for which we calculate the best performance
            instances: The instances which should be selected for the evaluation
            run_id: The run for which we calculate the best performance. If None,
                we consider all runs.
            exclude_solvers: List of (solver, config_id) to exclude in the calculation.

        Returns:
            The best performance for each instance in the portfolio.
        """
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        subdf = self.drop(  # Drop Seed, not needed
            [PerformanceDataFrame.column_seed],
            axis=1,
            level=PerformanceDataFrame.column_meta,
        )
        subdf = subdf.xs(objective.name, level=0)  # Drop objective
        if exclude_solvers is not None:
            subdf = subdf.drop(exclude_solvers, axis=1)
        if instances is not None:
            subdf = subdf.loc[instances, :]
        if run_id is not None:
            run_id = self.verify_run_id(run_id)
            subdf = subdf.xs(run_id, level=1)
        else:
            # Drop the run level
            subdf = subdf.droplevel(level=1)
        # Ensure the objective is numeric
        subdf = subdf.astype(float)
        series = subdf.min(axis=1) if objective.minimise else subdf.max(axis=1)
        # Ensure we always return the best for each run
        series = series.sort_values(ascending=objective.minimise)
        return series.groupby(series.index).first().astype(float)

    def best_performance(
        self: PerformanceDataFrame,
        exclude_solvers: list[(str, str)] = [],
        instances: list[str] = None,
        objective: str | SparkleObjective = None,
    ) -> float:
        """Return the overall best performance of the portfolio.

        Args:
            exclude_solvers: List of (solver, config_id) to exclude in the calculation.
                Defaults to none.
            instances: The instances which should be selected for the evaluation
                If None, use all instances.
            objective: The objective for which we calculate the best performance

        Returns:
            The aggregated best performance of the portfolio over all instances.
        """
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        instance_best = self.best_instance_performance(
            objective, instances=instances, exclude_solvers=exclude_solvers
        ).to_numpy(dtype=float)
        return objective.instance_aggregator(instance_best)

    def schedule_performance(
        self: PerformanceDataFrame,
        schedule: dict[str : dict[str : (str, str, int)]],
        target_solver: str | tuple[str, str] = None,
        objective: str | SparkleObjective = None,
    ) -> float:
        """Return the performance of a selection schedule on the portfolio.

        Args:
            schedule: Compute the best performance according to a selection schedule.
                A schedule is a dictionary of instances, with a schedule per instance,
                consisting of a triple of solver, config_id and maximum runtime.
            target_solver: If not None, store the found values in this solver of the DF.
            objective: The objective for which we calculate the best performance

        Returns:
            The performance of the schedule over the instances in the dictionary.
        """
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        select = min if objective.minimise else max
        performances = [0.0] * len(schedule.keys())
        if not isinstance(target_solver, tuple):
            target_conf = PerformanceDataFrame.default_configuration
        else:
            target_solver, target_conf = target_solver
        if target_solver and target_solver not in self.solvers:
            self.add_solver(target_solver)
        for ix, instance in enumerate(schedule.keys()):
            for iy, (solver, config, max_runtime) in enumerate(schedule[instance]):
                performance = float(
                    self.get_value(solver, instance, config, objective.name)
                )
                if max_runtime is not None:  # We are dealing with runtime
                    performances[ix] += performance
                    if performance < max_runtime:
                        break  # Solver finished in time
                else:  # Quality, we take the best found performance
                    if iy == 0:  # First solver, set initial value
                        performances[ix] = performance
                        continue
                    performances[ix] = select(performances[ix], performance)
            if target_solver is not None:
                self.set_value(
                    performances[ix],
                    target_solver,
                    instance,
                    target_conf,
                    objective.name,
                )
        return performances

    def marginal_contribution(
        self: PerformanceDataFrame,
        objective: str | SparkleObjective = None,
        instances: list[str] = None,
        sort: bool = False,
    ) -> list[float]:
        """Return the marginal contribution of the solver configuration on the instances.

        Args:
            objective: The objective for which we calculate the marginal contribution.
            instances: The instances which should be selected for the evaluation
            sort: Whether to sort the results afterwards
        Returns:
            The marginal contribution of each solver (configuration) as:
            [(solver, config_id, marginal_contribution, portfolio_best_performance_without_solver)]
        """
        output = []
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        best_performance = self.best_performance(
            objective=objective, instances=instances
        )
        for solver in self.solvers:
            for config_id in self.get_configurations(solver):
                # By calculating the best performance excluding this Solver,
                # we can determine its relative impact on the portfolio.
                missing_solver_config_best = self.best_performance(
                    exclude_solvers=[(solver, config_id)],
                    instances=instances,
                    objective=objective,
                )
                # Now we need to see how much the portfolio's best performance
                # decreases without this solver.
                marginal_contribution = missing_solver_config_best / best_performance
                if missing_solver_config_best == best_performance:
                    # No change, no contribution
                    marginal_contribution = 0.0
                output.append(
                    (
                        solver,
                        config_id,
                        marginal_contribution,
                        missing_solver_config_best,
                    )
                )
        if sort:
            output.sort(key=lambda x: x[2], reverse=objective.minimise)
        return output

    def get_solver_ranking(
        self: PerformanceDataFrame,
        objective: str | SparkleObjective = None,
        instances: list[str] = None,
    ) -> list[tuple[str, dict, float]]:
        """Return a list with solvers ranked by average performance."""
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        # Drop Seed
        sub_df = self.drop(
            [PerformanceDataFrame.column_seed],
            axis=1,
            level=PerformanceDataFrame.column_meta,
        )
        # Reduce objective
        sub_df: pd.DataFrame = sub_df.loc(axis=0)[objective.name, :, :]
        # Drop Objective, Meta multi index
        sub_df = sub_df.droplevel(PerformanceDataFrame.index_objective).droplevel(
            PerformanceDataFrame.column_meta, axis=1
        )
        if instances is not None:  # Select instances
            sub_df = sub_df.loc(axis=0)[instances,]
        # Ensure data is numeric
        sub_df = sub_df.astype(float)
        # Aggregate runs
        sub_df = sub_df.groupby(PerformanceDataFrame.index_instance).agg(
            func=objective.run_aggregator.__name__
        )
        # Aggregate instances
        sub_series = sub_df.aggregate(func=objective.instance_aggregator.__name__)
        # Sort by objective
        sub_series.sort_values(ascending=objective.minimise, inplace=True)
        return [(index[0], index[1], sub_series[index]) for index in sub_series.index]

    def save_csv(self: PerformanceDataFrame, csv_filepath: Path = None) -> None:
        """Write a CSV to the given path.

        Args:
            csv_filepath: String path to the csv file. Defaults to self.csv_filepath.
        """
        csv_filepath = self.csv_filepath if csv_filepath is None else csv_filepath
        self.to_csv(csv_filepath)
        # Append the configurations
        with csv_filepath.open("a") as fout:
            fout.write("\n$Solver,configuration_id,Configuration\n")
            for solver in self.solvers:
                for config_id in self.attrs[solver]:
                    configuration = self.attrs[solver][config_id]
                    fout.write(f"${solver},{config_id},{str(configuration)}\n")

    def clone(
        self: PerformanceDataFrame, csv_filepath: Path = None
    ) -> PerformanceDataFrame:
        """Create a copy of this object.

        Args:
            csv_filepath: The new filepath to use for saving the object to.
                If None, will not be saved.
                Warning: If the original path is used, it could lead to dataloss!
        """
        pd_copy = PerformanceDataFrame(
            csv_filepath=csv_filepath,
            solvers=self.solvers,
            configurations=self.configurations,
            objectives=self.objectives,
            instances=self.instances,
            n_runs=self.num_runs,
        )
        # Copy values
        for column_index in self.columns:
            for index in self.index:
                pd_copy.at[index, column_index] = self.loc[index, column_index]
        # Ensure everything is sorted?
        return pd_copy

    def clean_csv(self: PerformanceDataFrame) -> None:
        """Set all values in Performance Data to None."""
        self[:] = PerformanceDataFrame.missing_value
        self.save_csv()



#!/usr/bin/env python3
"""Sparkle command to execute a portfolio selector."""

import sys
import argparse

from runrunner import Runner

from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.platform.settings_objects import Settings
from sparkle.CLI.help import argparse_custom as ac
from sparkle.structures import FeatureDataFrame
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help.nicknames import resolve_object_name
from sparkle.instance import Instance_Set, InstanceSet
from sparkle.CLI.compute_features import compute_features
from sparkle.selector import SelectionScenario, Extractor


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Run a portfolio selector on instance (set): Determine which solver "
        "is most likely to perform well and run it on the instance (set)."
    )
    parser.add_argument(
        *ac.SelectionScenarioArgument.names, **ac.SelectionScenarioArgument.kwargs
    )
    parser.add_argument(
        *ac.InstanceSetRequiredArgument.names, **ac.InstanceSetRequiredArgument.kwargs
    )
    # Settings arguments
    parser.add_argument(*ac.SettingsFileArgument.names, **ac.SettingsFileArgument.kwargs)
    parser.add_argument(*Settings.OPTION_run_on.args, **Settings.OPTION_run_on.kwargs)
    return parser


def main(argv: list[str]) -> None:
    """Main function of the run portfolio selector command."""
    # Define command line arguments
    parser = parser_function()
    # Process command line arguments
    args = parser.parse_args(argv)
    settings = gv.settings(args)

    # Log command call
    sl.log_command(sys.argv, settings.random_state)
    check_for_initialise()

    # Compare current settings to latest.ini
    prev_settings = Settings(Settings.DEFAULT_previous_settings_path)
    Settings.check_settings_changes(settings, prev_settings)

    data_set: InstanceSet = resolve_object_name(
        args.instance,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        settings.DEFAULT_instance_dir,
        Instance_Set,
    )

    if data_set is None:
        print(
            "ERROR: The instance (set) could not be found. Please make sure the "
            "path is correct."
        )
        sys.exit(-1)

    run_on = settings.run_on
    selector_scenario = SelectionScenario.from_file(args.selection_scenario)
    # Create a new feature dataframe for this run, compute the features
    test_case_path = selector_scenario.directory / data_set.name
    test_case_path.mkdir(exist_ok=True)
    feature_dataframe = FeatureDataFrame(test_case_path / "feature_data.csv")
    feature_dataframe.remove_instances(feature_dataframe.instances)
    for extractor_name in selector_scenario.feature_extractors:
        extractor = resolve_object_name(
            extractor_name,
            gv.file_storage_data_mapping[gv.instances_nickname_path],
            settings.DEFAULT_extractor_dir,
            Extractor,
        )
        feature_dataframe.add_extractor(extractor_name, extractor.features)

    feature_dataframe.add_instances(data_set.instances)
    feature_dataframe.save_csv()
    feature_runs = compute_features(feature_dataframe, recompute=False, run_on=run_on)

    # Results need to be stored in the performance data object of the scenario:
    # Add the instance set to it
    for instance in data_set.instance_names:
        selector_scenario.selector_performance_data.add_instance(str(instance))
    selector_scenario.selector_performance_data.save_csv()

    selector_run = selector_scenario.selector.run_cli(
        scenario_path=selector_scenario.scenario_file,
        instance_set=data_set,
        feature_data=feature_dataframe.csv_filepath,
        run_on=run_on,
        slurm_prepend=settings.slurm_job_prepend,
        sbatch_options=settings.sbatch_settings,
        dependencies=feature_runs,
        log_dir=sl.caller_log_dir,
    )

    if run_on == Runner.LOCAL:
        selector_run.wait()
        print("Running Sparkle portfolio selector done!")
    else:
        print("Sparkle portfolio selector is running ...")

    # Write used settings to file
    settings.write_used_settings()
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



"""Methods regarding feature extractors."""

from __future__ import annotations
from typing import Any
from pathlib import Path
import ast
import subprocess

import runrunner as rrr
from runrunner.base import Status, Runner
from runrunner.local import Run, LocalRun

from sparkle.types import SparkleCallable, SolverStatus
from sparkle.structures import FeatureDataFrame
from sparkle.tools import RunSolver
from sparkle.instance import InstanceSet


class Extractor(SparkleCallable):
    """Extractor base class for extracting features from instances."""

    wrapper = "sparkle_extractor_wrapper.py"
    extractor_cli = Path(__file__).parent / "extractor_cli.py"

    def __init__(self: Extractor, directory: Path) -> None:
        """Initialize solver.

        Args:
            directory: Directory of the solver.
            runsolver_exec: Path to the runsolver executable.
                By default, runsolver in directory.
        """
        super().__init__(directory)
        self._features = None
        self._feature_groups = None
        self._groupwise_computation = None

    def __str__(self: Extractor) -> str:
        """Return the string representation of the extractor."""
        return self.name

    def __repr__(self: Extractor) -> str:
        """Return detailed representation of the extractor."""
        return (
            f"{self.name}:\n"
            f"\t- Directory: {self.directory}\n"
            f"\t- Wrapper: {self.wrapper}\n"
            f"\t- # Feature Groups: {len(self.feature_groups)}\n"
            f"\t- Output Dimension (# Features): {self.output_dimension}\n"
            f"\t- Groupwise Computation Enabled: {self.groupwise_computation}"
        )

    @property
    def features(self: Extractor) -> list[tuple[str, str]]:
        """Determines the features of the extractor."""
        if self._features is None:
            extractor_process = subprocess.run(
                [self.directory / Extractor.wrapper, "-features"], capture_output=True
            )
            self._features = ast.literal_eval(extractor_process.stdout.decode())
        return self._features

    @property
    def feature_groups(self: Extractor) -> list[str]:
        """Returns the various feature groups the Extractor has."""
        if self._feature_groups is None:
            self._feature_groups = list(set([group for group, _ in self.features]))
        return self._feature_groups

    @property
    def output_dimension(self: Extractor) -> int:
        """The size of the output vector of the extractor."""
        return len(self.features)

    @property
    def groupwise_computation(self: Extractor) -> bool:
        """Determines if you can call the extractor per group for parallelisation."""
        if self._groupwise_computation is None:
            extractor_help = subprocess.run(
                [self.directory / Extractor.wrapper, "-h"], capture_output=True
            )
            # Not the cleanest / most precise way to determine this
            self._groupwise_computation = (
                "-feature_group" in extractor_help.stdout.decode()
            )
        return self._groupwise_computation

    def build_cmd(
        self: Extractor,
        instance: Path | list[Path],
        feature_group: str = None,
        output_file: Path = None,
        cutoff_time: int = None,
        log_dir: Path = None,
    ) -> list[str]:
        """Builds a command line string seperated by space.

        Args:
            instance: The instance to run on
            feature_group: The optional feature group to run the extractor for.
            output_file: Optional file to write the output to.
            runsolver_args: The arguments for runsolver. If not present,
                will run the extractor without runsolver.
            cutoff_time: The maximum runtime.
            log_dir: Directory path for logs.

        Returns:
            The command seperated per item in the list.
        """
        cmd_list_extractor = []
        if not isinstance(instance, list):
            instance = [instance]
        cmd_list_extractor = [
            f"{self.directory / Extractor.wrapper}",
            "-extractor_dir",
            f"{self.directory}/",
            "-instance_file",
        ] + [str(file) for file in instance]
        if feature_group is not None:
            cmd_list_extractor += ["-feature_group", feature_group]
        if output_file is not None:
            cmd_list_extractor += ["-output_file", str(output_file)]
        if cutoff_time is not None:
            # Extractor handles output file itself
            return RunSolver.wrap_command(
                self.runsolver_exec,
                cmd_list_extractor,
                cutoff_time,
                log_dir,
                log_name_base=self.name,
                raw_results_file=False,
            )
        return cmd_list_extractor

    def run(
        self: Extractor,
        instance: Path | list[Path],
        feature_group: str = None,
        output_file: Path = None,
        cutoff_time: int = None,
        log_dir: Path = None,
    ) -> list[list[Any]] | list[Any] | None:
        """Runs an extractor job with Runrunner.

        Args:
            extractor_path: Path to the executable
            instance: Path to the instance to run on
            feature_group: The feature group to compute. Must be supported by the
                extractor to use.
            output_file: Target output. If None, piped to the RunRunner job.
            cutoff_time: CPU cutoff time in seconds
            log_dir: Directory to write logs. Defaults to CWD.

        Returns:
            The features or None if an output file is used, or features can not be found.
        """
        log_dir = Path() if log_dir is None else log_dir
        if feature_group is not None and not self.groupwise_computation:
            # This extractor cannot handle groups, compute all features
            feature_group = None
        cmd_extractor = self.build_cmd(
            instance, feature_group, output_file, cutoff_time, log_dir
        )
        run_on = Runner.LOCAL  # TODO: Let this function also handle Slurm runs
        extractor_run = rrr.add_to_queue(runner=run_on, cmd=" ".join(cmd_extractor))
        if isinstance(extractor_run, LocalRun):
            extractor_run.wait()
            if extractor_run.status == Status.ERROR:
                print(f"{self.name} failed to compute features for {instance}.")
                for i, job in enumerate(extractor_run.jobs):
                    print(
                        f"Job {i} error yielded was:\n"
                        f"\t-stdout: '{job.stdout}'\n"
                        f"\t-stderr: '{job.stderr}'\n"
                    )
                return None
            # RunRunner adds a timestamp before the statement
            import re

            pattern = re.compile(
                r"^(?P<timestamp1>\d+\.\d+)\/(?P<timestamp2>\d+\.\d+)\s+(?P<output>.*)$"
            )
            output = []
            for job in extractor_run.jobs:
                match = pattern.match(job.stdout)
                if match:
                    output.append(ast.literal_eval(match.group("output")))
            if len(output) == 1:
                return output[0]
            return output
        return None

    def run_cli(
        self: Extractor,
        instance_set: InstanceSet | list[Path],
        feature_dataframe: FeatureDataFrame,
        cutoff_time: int,
        feature_group: str = None,
        run_on: Runner = Runner.SLURM,
        sbatch_options: list[str] = None,
        srun_options: list[str] = None,
        parallel_jobs: int = None,
        slurm_prepend: str | list[str] | Path = None,
        dependencies: list[Run] = None,
        log_dir: Path = None,
    ) -> None:
        """Run the Extractor CLI and write result to the FeatureDataFrame.

        Args:
            instance_set: The instance set to run the Extractor on.
            feature_dataframe: The feature dataframe to write to.
            cutoff_time: CPU cutoff time in seconds
            feature_group: The feature group to compute. If left empty,
                will run on all feature groups.
            run_on: The runner to use.
            sbatch_options: Additional options to pass to sbatch.
            srun_options: Additional options to pass to srun.
            parallel_jobs: Number of parallel jobs to run.
            slurm_prepend: Slurm script to prepend to the sbatch
            dependencies: List of dependencies to add to the job.
            log_dir: The directory to write logs to.
        """
        instances = (
            instance_set
            if isinstance(instance_set, list)
            else instance_set.instance_paths
        )
        log_dir = Path() if log_dir is None else log_dir
        feature_group = f"--feature-group {feature_group} " if feature_group else ""
        commands = [
            f"python3 {Extractor.extractor_cli} "
            f"--extractor {self.directory} "
            f"--instance {instance_path} "
            f"--feature-csv {feature_dataframe.csv_filepath} "
            f"{feature_group}"
            f"--cutoff {cutoff_time} "
            f"--log-dir {log_dir}"
            for instance_path in instances
        ]

        job_name = f"Run Extractor: {self.name} on {len(instances)} instances"
        import subprocess

        run = rrr.add_to_queue(
            runner=run_on,
            cmd=commands,
            name=job_name,
            stdout=None if run_on == Runner.LOCAL else subprocess.PIPE,  # Print
            stderr=None if run_on == Runner.LOCAL else subprocess.PIPE,  # Print
            base_dir=log_dir,
            sbatch_options=sbatch_options,
            srun_options=srun_options,
            parallel_jobs=parallel_jobs,
            prepend=slurm_prepend,
            dependencies=dependencies,
        )
        if isinstance(run, LocalRun):
            print("Waiting for the local calculations to finish.")
            run.wait()
            for job in run.jobs:
                jobs_done = sum(j.status == Status.COMPLETED for j in run.jobs)
                print(f"Executing Progress: {jobs_done} out of {len(run.jobs)}")
                if jobs_done == len(run.jobs):
                    break
                job.wait()
            print("Computing features done!")
        else:
            print(f"Running {self.name} through Slurm with Job IDs: {run.run_id}")
        return run

    def get_feature_vector(
        self: Extractor, result: Path, runsolver_values: Path = None
    ) -> list[str]:
        """Extracts feature vector from an output file.

        Args:
            result: The raw output of the extractor
            runsolver_values: The output of runsolver.

        Returns:
            A list of features. Vector of missing values upon failure.
        """
        if (
            result.exists()
            and RunSolver.get_status(runsolver_values, None) != SolverStatus.TIMEOUT
        ):
            feature_values = ast.literal_eval(result.read_text())
            return [str(value) for _, _, value in feature_values]
        return [FeatureDataFrame.missing_value] * self.output_dimension
